#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Construct Hamiltonians in plan wave basis and its dual in 3D."""
from typing import List, Tuple, Optional, Union

import numpy as np

from openfermion.hamiltonians.jellium import (jellium_model,
                                              jordan_wigner_dual_basis_jellium)
from openfermion.ops.operators import FermionOperator, QubitOperator
from openfermion.transforms.repconversions import inverse_fourier_transform
from openfermion.utils.grid import Grid

import openfermion.chem.molecular_data as md


def dual_basis_external_potential(
        grid: Grid,
        geometry: List[Tuple[str, Tuple[Union[int, float], Union[int, float],
                                        Union[int, float]]]],
        spinless: bool,
        non_periodic: bool = False,
        period_cutoff: Optional[float] = None) -> FermionOperator:
    """Return the external potential in the dual basis of arXiv:1706.00023.

    The external potential resulting from electrons interacting with nuclei
        in the plane wave dual basis.  Note that a cos term is used which is
        strictly only equivalent under aliasing in odd grids, and amounts
        to the addition of an extra term to make the diagonals real on even
        grids.  This approximation is not expected to be significant and allows
        for use of even and odd grids on an even footing.

    Args:
        grid (Grid): The discretization to use.
        geometry: A list of tuples giving the coordinates of each atom.
            example is [('H', (0, 0, 0)), ('H', (0, 0, 0.7414))].
            Distances in atomic units. Use atomic symbols to specify atoms.
        spinless (bool): Whether to use the spinless model or not.
        non_periodic (bool): If the system is non-periodic, default to False.
        period_cutoff (float): Period cutoff, default to
            grid.volume_scale() ** (1. / grid.dimensions)

    Returns:
        FermionOperator: The dual basis operator.
    """
    prefactor = -4.0 * np.pi / grid.volume_scale()
    if non_periodic and period_cutoff is None:
        period_cutoff = grid.volume_scale()**(1. / grid.dimensions)
    operator = None
    if spinless:
        spins = [None]
    else:
        spins = [0, 1]
    for pos_indices in grid.all_points_indices():
        coordinate_p = grid.position_vector(pos_indices)
        for nuclear_term in geometry:
            coordinate_j = np.array(nuclear_term[1], float)
            for momenta_indices in grid.all_points_indices():
                momenta = grid.momentum_vector(momenta_indices)
                momenta_squared = momenta.dot(momenta)
                if momenta_squared == 0:
                    continue

                cos_index = momenta.dot(coordinate_j - coordinate_p)
                coefficient = (prefactor / momenta_squared *
                               md.periodic_hash_table[nuclear_term[0]] *
                               np.cos(cos_index))

                for spin_p in spins:
                    orbital_p = grid.orbital_id(pos_indices, spin_p)
                    operators = ((orbital_p, 1), (orbital_p, 0))
                    if operator is None:
                        operator = FermionOperator(operators, coefficient)
                    else:
                        operator += FermionOperator(operators, coefficient)
    return operator


def plane_wave_external_potential(
        grid: Grid,
        geometry: List[Tuple[str, Tuple[Union[int, float], Union[int, float],
                                        Union[int, float]]]],
        spinless: bool,
        e_cutoff: Optional[float] = None,
        non_periodic: bool = False,
        period_cutoff: Optional[float] = None) -> FermionOperator:
    """Return the external potential operator in plane wave basis.

    The external potential resulting from electrons interacting with nuclei.
        It is defined here as the Fourier transform of the dual basis
        Hamiltonian such that is spectrally equivalent in the case of
        both even and odd grids.  Otherwise, the two differ in the case of
        even grids.

    Args:
        grid (Grid): The discretization to use.
        geometry: A list of tuples giving the coordinates of each atom.
            example is [('H', (0, 0, 0)), ('H', (0, 0, 0.7414))].
            Distances in atomic units. Use atomic symbols to specify atoms.
        spinless: Bool, whether to use the spinless model or not.
        e_cutoff (float): Energy cutoff.
        non_periodic (bool): If the system is non-periodic, default to False.
        period_cutoff (float): Period cutoff, default to
            grid.volume_scale() ** (1. / grid.dimensions)

    Returns:
        FermionOperator: The plane wave operator.
    """
    dual_basis_operator = dual_basis_external_potential(grid, geometry,
                                                        spinless, non_periodic,
                                                        period_cutoff)
    operator = inverse_fourier_transform(dual_basis_operator, grid, spinless)

    return operator


def plane_wave_hamiltonian(
        grid: Grid,
        geometry: Optional[List[Tuple[str, Tuple[
            Union[int, float], Union[int, float], Union[int, float]]]]] = None,
        spinless: bool = False,
        plane_wave: bool = True,
        include_constant: bool = False,
        e_cutoff: Optional[float] = None,
        non_periodic: bool = False,
        period_cutoff: Optional[float] = None) -> FermionOperator:
    """Returns Hamiltonian as FermionOperator class.

    Args:
        grid (Grid): The discretization to use.
        geometry: A list of tuples giving the coordinates of each atom.
            example is [('H', (0, 0, 0)), ('H', (0, 0, 0.7414))].
            Distances in atomic units. Use atomic symbols to specify atoms.
        spinless (bool): Whether to use the spinless model or not.
        plane_wave (bool): Whether to return in plane wave basis (True)
            or plane wave dual basis (False).
        include_constant (bool): Whether to include the Madelung constant.
        e_cutoff (float): Energy cutoff.
        non_periodic (bool): If the system is non-periodic, default to False.
        period_cutoff (float): Period cutoff, default to
            grid.volume_scale() ** (1. / grid.dimensions)

    Returns:
        FermionOperator: The hamiltonian.
    """
    if (geometry is not None) and (include_constant is True):
        raise ValueError('Constant term unsupported for non-uniform systems')

    jellium_op = jellium_model(grid, spinless, plane_wave, include_constant,
                               e_cutoff, non_periodic, period_cutoff)

    if geometry is None:
        return jellium_op

    for item in geometry:
        if len(item[1]) != grid.dimensions:
            raise ValueError("Invalid geometry coordinate.")
        if item[0] not in md.periodic_hash_table:
            raise ValueError("Invalid nuclear element.")

    if plane_wave:
        external_potential = plane_wave_external_potential(
            grid, geometry, spinless, e_cutoff, non_periodic, period_cutoff)
    else:
        external_potential = dual_basis_external_potential(
            grid, geometry, spinless, non_periodic, period_cutoff)

    return jellium_op + external_potential


def jordan_wigner_dual_basis_hamiltonian(
        grid: Grid,
        geometry: Optional[List[Tuple[str, Tuple[
            Union[int, float], Union[int, float], Union[int, float]]]]] = None,
        spinless: bool = False,
        include_constant: bool = False) -> QubitOperator:
    """Return the dual basis Hamiltonian as QubitOperator.

    Args:
        grid (Grid): The discretization to use.
        geometry: A list of tuples giving the coordinates of each atom.
            example is [('H', (0, 0, 0)), ('H', (0, 0, 0.7414))].
            Distances in atomic units. Use atomic symbols to specify atoms.
        spinless (bool): Whether to use the spinless model or not.
        include_constant (bool): Whether to include the Madelung constant.

    Returns:
        hamiltonian (QubitOperator)
    """
    if (geometry is not None) and (include_constant is True):
        raise ValueError('Constant term unsupported for non-uniform systems')

    jellium_op = jordan_wigner_dual_basis_jellium(grid, spinless,
                                                  include_constant)

    if geometry is None:
        return jellium_op

    for item in geometry:
        if len(item[1]) != grid.dimensions:
            raise ValueError("Invalid geometry coordinate.")
        if item[0] not in md.periodic_hash_table:
            raise ValueError("Invalid nuclear element.")

    n_orbitals = grid.num_points
    volume = grid.volume_scale()
    if spinless:
        n_qubits = n_orbitals
    else:
        n_qubits = 2 * n_orbitals
    prefactor = -2 * np.pi / volume
    external_potential = QubitOperator()

    for k_indices in grid.all_points_indices():
        momenta = grid.momentum_vector(k_indices)
        momenta_squared = momenta.dot(momenta)
        if momenta_squared == 0:
            continue

        for p in range(n_qubits):
            index_p = grid.grid_indices(p, spinless)
            coordinate_p = grid.position_vector(index_p)

            for nuclear_term in geometry:
                coordinate_j = np.array(nuclear_term[1], float)

                cos_index = momenta.dot(coordinate_j - coordinate_p)
                coefficient = (prefactor / momenta_squared *
                               md.periodic_hash_table[nuclear_term[0]] *
                               np.cos(cos_index))
                external_potential += (QubitOperator(
                    (), coefficient) - QubitOperator(((p, 'Z'),), coefficient))

    return jellium_op + external_potential



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import pytest
import numpy
import cirq
import openfermion
from openfermion.ops.operators import QubitOperator

from openfermion.transforms.opconversions import (qubit_operator_to_pauli_sum)
from openfermion.transforms.opconversions.qubitoperator_to_paulisum import (
    _qubit_operator_term_to_pauli_string)


def test_function_raises():
    """Test function raises."""
    operator = QubitOperator('X0 X1 X2 X3', 1.0)
    with pytest.raises(TypeError):
        _qubit_operator_term_to_pauli_string(list(operator.terms.items())[0],
                                             qubits=0.0)
    with pytest.raises(TypeError):
        qubit_operator_to_pauli_sum([5.0])


def test_pauli_sum_qubits():
    """Test PauliSum creates qubits."""
    operator = QubitOperator('X0 X1 X2 X3', 1.0)
    pauli_sum = qubit_operator_to_pauli_sum(operator=operator)

    assert list(pauli_sum.qubits) == (cirq.LineQubit.range(4))


def test_identity():
    """Test correct hanlding of Identity."""
    identity_op = QubitOperator(' ', -0.5)
    pau_from_qop = _qubit_operator_term_to_pauli_string(
        term=list(identity_op.terms.items())[0], qubits=cirq.LineQubit.range(2))
    pauli_str = cirq.PauliString() * (-0.5)

    assert pauli_str == pau_from_qop


@pytest.mark.parametrize('qubitop, state_binary',
                         [(QubitOperator('Z0 Z1', -1.0), '00'),
                          (QubitOperator('X0 Y1', 1.0), '10')])
def test_expectation_values(qubitop, state_binary):
    """Test PauliSum and QubitOperator expectation value."""
    n_qubits = openfermion.count_qubits(qubitop)
    state = numpy.zeros(2**n_qubits, dtype='complex64')
    state[int(state_binary, 2)] = 1.0
    qubit_map = {cirq.LineQubit(i): i for i in range(n_qubits)}

    pauli_str = _qubit_operator_term_to_pauli_string(
        term=list(qubitop.terms.items())[0],
        qubits=cirq.LineQubit.range(n_qubits))
    op_mat = openfermion.get_sparse_operator(qubitop, n_qubits)

    expct_qop = openfermion.expectation(op_mat, state)
    expct_pauli = pauli_str.expectation_from_state_vector(state, qubit_map)

    numpy.testing.assert_allclose(expct_qop, expct_pauli)


@pytest.mark.parametrize(
    'qubitop, state_binary',
    [(QubitOperator('Z0 Z1 Z2 Z3', -1.0) + QubitOperator('X0 Y1 Y2 X3', 1.0),
      '1100'),
     (QubitOperator('X0 X3', -1.0) + QubitOperator('Y1 Y2', 1.0), '0000')])
def test_expectation_values_paulisum(qubitop, state_binary):
    """Test PauliSum and QubitOperator expectation value."""
    n_qubits = openfermion.count_qubits(qubitop)
    state = numpy.zeros(2**n_qubits, dtype='complex64')
    state[int(state_binary, 2)] = 1.0
    qubit_map = {cirq.LineQubit(i): i for i in range(n_qubits)}

    pauli_str = qubit_operator_to_pauli_sum(qubitop, list(qubit_map.keys()))
    op_mat = openfermion.get_sparse_operator(qubitop, n_qubits)

    expct_qop = openfermion.expectation(op_mat, state)
    expct_pauli = pauli_str.expectation_from_state_vector(state, qubit_map)

    numpy.testing.assert_allclose(expct_qop, expct_pauli)



from itertools import product
from openfermion.third_party.representability._bijections import Bijection, \
    index_index_basis, index_tuple_basis


def test_bijection():
    """
    Testing the basis bijection
    """
    b = Bijection(lambda x: x + 1, lambda y: y - 1, lambda: (1, 1))
    assert isinstance(b, Bijection)
    assert b.fwd(2) == 3
    assert b.rev(2) == 1
    assert b.fwd(b.rev(5)) == 5
    assert b.domain_element_sizes() == (1, 1)


def test_index_basis():
    b = index_index_basis(5)
    assert b.fwd(4) == 4
    assert b.rev(4) == 4
    assert b.domain_element_sizes() == (1, 1)


def test_geminal_basis():
    gems = list(product(range(5), repeat=2))
    b = index_tuple_basis(gems)
    assert b.fwd(4) == (0, 4)
    assert b.rev((0, 4)) == 4
    assert b.rev(b.fwd(4)) == 4
    assert b.domain_element_sizes() == (1, 2)


#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module constructs Hamiltonians for the uniform electron gas."""

from typing import Optional

import numpy

from openfermion.ops.operators import FermionOperator, QubitOperator
from openfermion.utils.grid import Grid


def wigner_seitz_length_scale(wigner_seitz_radius: float, n_particles: int,
                              dimension: int) -> float:
    """Function to give length_scale associated with Wigner-Seitz radius.

    Args:
        wigner_seitz_radius (float): The radius per particle in atomic units.
        n_particles (int): The number of particles in the simulation cell.
        dimension (int): The dimension of the system.

    Returns:
        length_scale (float): The length scale for the simulation.

    Raises:
        ValueError: System dimension must be a positive integer.
    """
    if not isinstance(dimension, int) or dimension < 1:
        raise ValueError('System dimension must be a positive integer.')

    half_dimension = dimension // 2
    if dimension % 2:
        volume_per_particle = (2 * numpy.math.factorial(half_dimension) *
                               (4 * numpy.pi)**half_dimension /
                               numpy.math.factorial(dimension) *
                               wigner_seitz_radius**dimension)
    else:
        volume_per_particle = (numpy.pi**half_dimension /
                               numpy.math.factorial(half_dimension) *
                               wigner_seitz_radius**dimension)

    volume = volume_per_particle * n_particles
    length_scale = volume**(1. / dimension)

    return length_scale


def plane_wave_kinetic(grid: Grid,
                       spinless: bool = False,
                       e_cutoff: Optional[float] = None) -> FermionOperator:
    """Return the kinetic energy operator in the plane wave basis.

    Args:
        grid (openfermion.utils.Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.
        e_cutoff (float): Energy cutoff.

    Returns:
        FermionOperator: The kinetic momentum operator.
    """
    # Initialize.
    operator = FermionOperator()
    spins = [None] if spinless else [0, 1]

    # Loop once through all plane waves.
    for momenta_indices in grid.all_points_indices():
        momenta = grid.momentum_vector(momenta_indices)
        coefficient = momenta.dot(momenta) / 2.

        # Energy cutoff.
        if e_cutoff is not None and coefficient > e_cutoff:
            continue

        # Loop over spins.
        for spin in spins:
            orbital = grid.orbital_id(momenta_indices, spin)

            # Add interaction term.
            operators = ((orbital, 1), (orbital, 0))
            operator += FermionOperator(operators, coefficient)

    return operator


def plane_wave_potential(grid: Grid,
                         spinless: bool = False,
                         e_cutoff: float = None,
                         non_periodic: bool = False,
                         period_cutoff: Optional[float] = None
                        ) -> FermionOperator:
    """Return the e-e potential operator in the plane wave basis.

    Args:
        grid (Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.
        e_cutoff (float): Energy cutoff.
        non_periodic (bool): If the system is non-periodic, default to False.
        period_cutoff (float): Period cutoff, default to
            grid.volume_scale() ** (1. / grid.dimensions).

    Returns:
        operator (FermionOperator)
    """
    # Initialize.
    prefactor = 2. * numpy.pi / grid.volume_scale()
    operator = FermionOperator((), 0.0)
    spins = [None] if spinless else [0, 1]
    if non_periodic and period_cutoff is None:
        period_cutoff = grid.volume_scale()**(1. / grid.dimensions)

    # Pre-Computations.
    shifted_omega_indices_dict = {}
    shifted_indices_minus_dict = {}
    shifted_indices_plus_dict = {}
    orbital_ids = {}
    for indices_a in grid.all_points_indices():
        shifted_omega_indices = [
            j - grid.length[i] // 2 for i, j in enumerate(indices_a)
        ]
        shifted_omega_indices_dict[indices_a] = shifted_omega_indices
        shifted_indices_minus_dict[indices_a] = {}
        shifted_indices_plus_dict[indices_a] = {}
        for indices_b in grid.all_points_indices():
            shifted_indices_minus_dict[indices_a][indices_b] = tuple([
                (indices_b[i] - shifted_omega_indices[i]) % grid.length[i]
                for i in range(grid.dimensions)
            ])
            shifted_indices_plus_dict[indices_a][indices_b] = tuple([
                (indices_b[i] + shifted_omega_indices[i]) % grid.length[i]
                for i in range(grid.dimensions)
            ])
        orbital_ids[indices_a] = {}
        for spin in spins:
            orbital_ids[indices_a][spin] = grid.orbital_id(indices_a, spin)

    # Loop once through all plane waves.
    for omega_indices in grid.all_points_indices():
        shifted_omega_indices = shifted_omega_indices_dict[omega_indices]

        # Get the momenta vectors.
        momenta = grid.momentum_vector(omega_indices)
        momenta_squared = momenta.dot(momenta)

        # Skip if momentum is zero.
        if momenta_squared == 0:
            continue

        # Energy cutoff.
        if e_cutoff is not None and momenta_squared / 2. > e_cutoff:
            continue

        # Compute coefficient.
        coefficient = prefactor / momenta_squared
        if non_periodic:
            coefficient *= 1.0 - numpy.cos(
                period_cutoff * numpy.sqrt(momenta_squared))

        for grid_indices_a in grid.all_points_indices():
            shifted_indices_d = (
                shifted_indices_minus_dict[omega_indices][grid_indices_a])
            for grid_indices_b in grid.all_points_indices():
                shifted_indices_c = (
                    shifted_indices_plus_dict[omega_indices][grid_indices_b])

                # Loop over spins.
                for spin_a in spins:
                    orbital_a = orbital_ids[grid_indices_a][spin_a]
                    orbital_d = orbital_ids[shifted_indices_d][spin_a]
                    for spin_b in spins:
                        orbital_b = orbital_ids[grid_indices_b][spin_b]
                        orbital_c = orbital_ids[shifted_indices_c][spin_b]

                        # Add interaction term.
                        if ((orbital_a != orbital_b) and
                            (orbital_c != orbital_d)):
                            operators = ((orbital_a, 1), (orbital_b, 1),
                                         (orbital_c, 0), (orbital_d, 0))
                            operator += FermionOperator(operators, coefficient)

    # Return.
    return operator


def dual_basis_jellium_model(grid: Grid,
                             spinless: bool = False,
                             kinetic: bool = True,
                             potential: bool = True,
                             include_constant: bool = False,
                             non_periodic: bool = False,
                             period_cutoff: Optional[float] = None
                            ) -> FermionOperator:
    """Return jellium Hamiltonian in the dual basis of arXiv:1706.00023

    Args:
        grid (Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.
        kinetic (bool): Whether to include kinetic terms.
        potential (bool): Whether to include potential terms.
        include_constant (bool): Whether to include the Madelung constant.
            Note constant is unsupported for non-uniform, non-cubic cells with
            ions.
        non_periodic (bool): If the system is non-periodic, default to False.
        period_cutoff (float): Period cutoff, default to
            grid.volume_scale() ** (1. / grid.dimensions).

    Returns:
        operator (FermionOperator)
    """
    # Initialize.
    n_points = grid.num_points
    position_prefactor = 2.0 * numpy.pi / grid.volume_scale()
    operator = FermionOperator()
    spins = [None] if spinless else [0, 1]
    if potential and non_periodic and period_cutoff is None:
        period_cutoff = grid.volume_scale()**(1.0 / grid.dimensions)

    # Pre-Computations.
    position_vectors = {}
    momentum_vectors = {}
    momenta_squared_dict = {}
    orbital_ids = {}
    for indices in grid.all_points_indices():
        position_vectors[indices] = grid.position_vector(indices)
        momenta = grid.momentum_vector(indices)
        momentum_vectors[indices] = momenta
        momenta_squared_dict[indices] = momenta.dot(momenta)
        orbital_ids[indices] = {}
        for spin in spins:
            orbital_ids[indices][spin] = grid.orbital_id(indices, spin)

    # Loop once through all lattice sites.
    grid_origin = (0,) * grid.dimensions
    coordinates_origin = position_vectors[grid_origin]
    for grid_indices_b in grid.all_points_indices():
        coordinates_b = position_vectors[grid_indices_b]
        differences = coordinates_b - coordinates_origin

        # Compute coefficients.
        kinetic_coefficient = 0.
        potential_coefficient = 0.
        for momenta_indices in grid.all_points_indices():
            momenta = momentum_vectors[momenta_indices]
            momenta_squared = momenta_squared_dict[momenta_indices]
            if momenta_squared == 0:
                continue

            cos_difference = numpy.cos(momenta.dot(differences))
            if kinetic:
                kinetic_coefficient += (cos_difference * momenta_squared /
                                        (2. * float(n_points)))
            if potential:
                potential_coefficient += (position_prefactor * cos_difference /
                                          momenta_squared)
        for grid_indices_shift in grid.all_points_indices():
            # Loop over spins and identify interacting orbitals.
            orbital_a = {}
            orbital_b = {}
            shifted_index_1 = tuple([
                (grid_origin[i] + grid_indices_shift[i]) % grid.length[i]
                for i in range(grid.dimensions)
            ])
            shifted_index_2 = tuple([
                (grid_indices_b[i] + grid_indices_shift[i]) % grid.length[i]
                for i in range(grid.dimensions)
            ])

            for spin in spins:
                orbital_a[spin] = orbital_ids[shifted_index_1][spin]
                orbital_b[spin] = orbital_ids[shifted_index_2][spin]
            if kinetic:
                for spin in spins:
                    operators = ((orbital_a[spin], 1), (orbital_b[spin], 0))
                    operator += FermionOperator(operators, kinetic_coefficient)
            if potential:
                for sa in spins:
                    for sb in spins:
                        if orbital_a[sa] == orbital_b[sb]:
                            continue
                        operators = ((orbital_a[sa], 1), (orbital_a[sa], 0),
                                     (orbital_b[sb], 1), (orbital_b[sb], 0))
                        operator += FermionOperator(operators,
                                                    potential_coefficient)

    # Include the Madelung constant if requested.
    if include_constant:
        # TODO: Check for other unit cell shapes
        operator += (FermionOperator.identity() *
                     (2.8372 / grid.volume_scale()**(1. / grid.dimensions)))

    # Return.
    return operator


def dual_basis_kinetic(grid: Grid, spinless: bool = False) -> FermionOperator:
    """Return the kinetic operator in the dual basis of arXiv:1706.00023.

    Args:
        grid (Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.

    Returns:
        operator (FermionOperator)
    """
    return dual_basis_jellium_model(grid, spinless, True, False)


def dual_basis_potential(grid: Grid,
                         spinless: bool = False,
                         non_periodic: bool = False,
                         period_cutoff: Optional[float] = None
                        ) -> FermionOperator:
    """Return the potential operator in the dual basis of arXiv:1706.00023

    Args:
        grid (Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.
        non_periodic (bool): If the system is non-periodic, default to False.
        period_cutoff (float): Period cutoff, default to
            grid.volume_scale() ** (1. / grid.dimensions).

    Returns:
        operator (FermionOperator)
    """
    return dual_basis_jellium_model(grid, spinless, False, True, False,
                                    non_periodic, period_cutoff)


def jellium_model(grid: Grid,
                  spinless: bool = False,
                  plane_wave: bool = True,
                  include_constant: bool = False,
                  e_cutoff: float = None,
                  non_periodic: bool = False,
                  period_cutoff: Optional[float] = None) -> FermionOperator:
    """Return jellium Hamiltonian as FermionOperator class.

    Args:
        grid (openfermion.utils.Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.
        plane_wave (bool): Whether to return in momentum space (True)
            or position space (False).
        include_constant (bool): Whether to include the Madelung constant.
            Note constant is unsupported for non-uniform, non-cubic cells with
            ions.
        e_cutoff (float): Energy cutoff.
        non_periodic (bool): If the system is non-periodic, default to False.
        period_cutoff (float): Period cutoff, default to
            grid.volume_scale() ** (1. / grid.dimensions).

    Returns:
        FermionOperator: The Hamiltonian of the model.
    """
    if plane_wave:
        hamiltonian = plane_wave_kinetic(grid, spinless, e_cutoff)
        hamiltonian += plane_wave_potential(grid, spinless, e_cutoff,
                                            non_periodic, period_cutoff)
    else:
        hamiltonian = dual_basis_jellium_model(grid, spinless, True, True,
                                               include_constant, non_periodic,
                                               period_cutoff)
    # Include the Madelung constant if requested.
    if include_constant:
        # TODO: Check for other unit cell shapes
        hamiltonian += (FermionOperator.identity() *
                        (2.8372 / grid.volume_scale()**(1. / grid.dimensions)))
    return hamiltonian


def jordan_wigner_dual_basis_jellium(grid: Grid,
                                     spinless: bool = False,
                                     include_constant: bool = False
                                    ) -> QubitOperator:
    """Return the jellium Hamiltonian as QubitOperator in the dual basis.

    Args:
        grid (Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.
        include_constant (bool): Whether to include the Madelung constant.
            Note constant is unsupported for non-uniform, non-cubic cells with
            ions.

    Returns:
        hamiltonian (QubitOperator)
    """
    # Initialize.
    n_orbitals = grid.num_points
    volume = grid.volume_scale()
    if spinless:
        n_qubits = n_orbitals
    else:
        n_qubits = 2 * n_orbitals
    hamiltonian = QubitOperator()

    # Compute vectors.
    momentum_vectors = {}
    momenta_squared_dict = {}
    for indices in grid.all_points_indices():
        momenta = grid.momentum_vector(indices)
        momentum_vectors[indices] = momenta
        momenta_squared_dict[indices] = momenta.dot(momenta)

    # Compute the identity coefficient and the coefficient of local Z terms.
    identity_coefficient = 0.
    z_coefficient = 0.
    for k_indices in grid.all_points_indices():
        momenta = momentum_vectors[k_indices]
        momenta_squared = momenta.dot(momenta)
        if momenta_squared == 0:
            continue

        identity_coefficient += momenta_squared / 2.
        identity_coefficient -= (numpy.pi * float(n_orbitals) /
                                 (momenta_squared * volume))
        z_coefficient += numpy.pi / (momenta_squared * volume)
        z_coefficient -= momenta_squared / (4. * float(n_orbitals))
    if spinless:
        identity_coefficient /= 2.

    # Add identity term.
    identity_term = QubitOperator((), identity_coefficient)
    hamiltonian += identity_term

    # Add local Z terms.
    for qubit in range(n_qubits):
        qubit_term = QubitOperator(((qubit, 'Z'),), z_coefficient)
        hamiltonian += qubit_term

    # Add ZZ terms and XZX + YZY terms.
    zz_prefactor = numpy.pi / volume
    xzx_yzy_prefactor = .25 / float(n_orbitals)
    for p in range(n_qubits):
        index_p = grid.grid_indices(p, spinless)
        position_p = grid.position_vector(index_p)
        for q in range(p + 1, n_qubits):
            index_q = grid.grid_indices(q, spinless)
            position_q = grid.position_vector(index_q)

            difference = position_p - position_q

            skip_xzx_yzy = not spinless and (p + q) % 2

            # Loop through momenta.
            zpzq_coefficient = 0.
            term_coefficient = 0.
            for k_indices in grid.all_points_indices():
                momenta = momentum_vectors[k_indices]
                momenta_squared = momenta_squared_dict[k_indices]
                if momenta_squared == 0:
                    continue

                cos_difference = numpy.cos(momenta.dot(difference))

                zpzq_coefficient += (zz_prefactor * cos_difference /
                                     momenta_squared)

                if skip_xzx_yzy:
                    continue
                term_coefficient += (xzx_yzy_prefactor * cos_difference *
                                     momenta_squared)

            # Add ZZ term.
            qubit_term = QubitOperator(((p, 'Z'), (q, 'Z')), zpzq_coefficient)
            hamiltonian += qubit_term

            # Add XZX + YZY term.
            if skip_xzx_yzy:
                continue
            z_string = tuple((i, 'Z') for i in range(p + 1, q))
            xzx_operators = ((p, 'X'),) + z_string + ((q, 'X'),)
            yzy_operators = ((p, 'Y'),) + z_string + ((q, 'Y'),)
            hamiltonian += QubitOperator(xzx_operators, term_coefficient)
            hamiltonian += QubitOperator(yzy_operators, term_coefficient)

    # Include the Madelung constant if requested.
    if include_constant:
        # TODO Generalize to other cells
        hamiltonian += (QubitOperator(
            (),) * (2.8372 / grid.volume_scale()**(1. / grid.dimensions)))

    # Return Hamiltonian.
    return hamiltonian


def hypercube_grid_with_given_wigner_seitz_radius_and_filling(
        dimension: int,
        grid_length: int,
        wigner_seitz_radius: float,
        filling_fraction: float = 0.5,
        spinless: bool = True) -> Grid:
    """Return a Grid with the same number of orbitals along each dimension
    with the specified Wigner-Seitz radius.

    Args:
        dimension (int): The number of spatial dimensions.
        grid_length (int): The number of orbitals along each dimension.
        wigner_seitz_radius (float): The Wigner-Seitz radius per particle,
            in Bohr.
        filling_fraction (float): The average spin-orbital occupation.
            Specifies the number of particles (rounding down).
        spinless (boolean): Whether to give the system without or with spin.
    """
    if filling_fraction > 1:
        raise ValueError("filling_fraction cannot be greater than 1.")

    n_qubits = grid_length**dimension
    if not spinless:
        n_qubits *= 2

    n_particles = int(numpy.floor(n_qubits * filling_fraction))

    if not n_particles:
        raise ValueError(
            "filling_fraction too low for number of orbitals specified by "
            "other parameters.")

    # Compute appropriate length scale.
    length_scale = wigner_seitz_length_scale(wigner_seitz_radius, n_particles,
                                             dimension)

    return Grid(dimension, grid_length, length_scale)



from typing import List, Optional, Union
from itertools import product
from openfermion.third_party.representability._dualbasis import \
    DualBasisElement, DualBasis
from openfermion.utils.rdm_mapping_functions import kronecker_delta


def tpdm_trace_constraint(dim: int, normalization: float) -> DualBasisElement:
    """
    Generate the trace constraint on the 2-PDM

    Args:
        dim:  Dimension of the spin-orbital single-particle Hilbert space
        normalization: Desired trace value.

    Returns: A DualBasisElement (Row of the constraint matrix) encoding
             the trace constraint on the 2-RDM.
    """
    tensor_elements = [(i, j, i, j) for i, j in product(range(dim), repeat=2)]
    tensor_names = ['cckk'] * (dim**2)
    tensor_coeffs = [1.0] * (dim**2)
    bias = 0
    return DualBasisElement(tensor_names=tensor_names,
                            tensor_elements=tensor_elements,
                            tensor_coeffs=tensor_coeffs,
                            bias=bias,
                            scalar=normalization)


def tpdm_antisymmetry_constraint(dim: int) -> DualBasis:
    """
    The dual basis elements representing the antisymmetry constraints

    :param dim: spinless Fermion basis rank
    :return: the dual basis of antisymmetry_constraints
    :rtype: DualBasis
    """
    # dual_basis = DualBasis()
    dbe_list = []
    for p, q, r, s in product(range(dim), repeat=4):
        if p * dim + q <= r * dim + s:
            if p < q and r < s:
                tensor_elements = [
                    tuple(indices) for indices in _coord_generator(p, q, r, s)
                ]
                tensor_names = ['cckk'] * len(tensor_elements)
                tensor_coeffs = [0.5] * len(tensor_elements)
                dbe = DualBasisElement()
                for n, e, c in zip(tensor_names, tensor_elements,
                                   tensor_coeffs):
                    dbe.add_element(n, e, c)

                # dual_basis += dbe
                dbe_list.append(dbe)

    return DualBasis(elements=dbe_list)


def tpdm_to_opdm_mapping(dim: int,
                         normalization: Union[float, int]) -> DualBasis:
    """
    Construct the DualBasis for mapping of the tpdm to the opdm

    Args:
        dim: dimension of the spin-orbital basis.
        normalization: Scalar for mapping tpdm to opdm.  Generally, this is
                       1 / (N - 1) where N is the number of electrons.
    Returns:
        DualBasis for all dim^2
    """
    db_basis = DualBasis()
    for i in range(dim):
        for j in range(i, dim):
            dbe = DualBasisElement()
            # contract over tpdm terms
            for r in range(dim):
                # duplicate entries get summed in DualBasisElement
                dbe.add_element('cckk', (i, r, j, r), 0.5)
                dbe.add_element('cckk', (j, r, i, r), 0.5)

            # opdm terms
            dbe.add_element('ck', (i, j), -0.5 * normalization)
            dbe.add_element('ck', (j, i), -0.5 * normalization)
            dbe.simplify()
            db_basis += dbe

    return db_basis


def opdm_to_ohdm_mapping(dim: int) -> DualBasis:
    """
    Map the ck to kc

    D1 + Q1 = I

    Args:
        dim: dimension of the spin-orbital basis
    Returns:
        DualBasis for the 1-RDM representability constraint
    """
    dbe_list = []
    for i in range(dim):
        for j in range(i, dim):
            dbe = DualBasisElement()
            if i != j:
                dbe.add_element('ck', (i, j), 0.5)
                dbe.add_element('ck', (j, i), 0.5)
                dbe.add_element('kc', (j, i), 0.5)
                dbe.add_element('kc', (i, j), 0.5)
                dbe.dual_scalar = 0.0
            else:
                dbe.add_element('ck', (i, j), 1.0)
                dbe.add_element('kc', (i, j), 1.0)
                dbe.dual_scalar = 1.0

            # db += dbe
            dbe_list.append(dbe)

    return DualBasis(elements=dbe_list)  # db


def sz_constraint(dim: int, sz: Union[float, int]) -> DualBasis:
    """
    Constraint on the 1-RDM

    Args:
        dim: dimension of the spin-orbital basis.
        sz: expectation value of the magnetic quantum number.

    Returns:
        DualBasis
    """
    dbe = DualBasisElement()
    for i in range(dim // 2):
        dbe.add_element('ck', (2 * i, 2 * i), 0.5)
        dbe.add_element('ck', (2 * i + 1, 2 * i + 1), -0.5)
    dbe.dual_scalar = sz
    return DualBasis(elements=[dbe])


def na_constraint(dim: int, na: Union[float, int]) -> DualBasis:
    """
    Constraint the trace of the alpha block of the opdm to equal the number
    of spin-up electrons

    Args:
        dim: Dimension of the spin-orbital basis.
        na: Number of spin up electrons

    Returns:
        DualBasis representing the cosntraint that is length 1
    """

    dbe = DualBasisElement()
    for i in range(dim // 2):
        dbe.add_element('ck', (2 * i, 2 * i), 1.0)
    dbe.dual_scalar = na
    return DualBasis(elements=[dbe])


def nb_constraint(dim, nb):
    """
    Constraint the trace of the alpha block of the opdm to equal the number
    of spin-down electrons

    Args:
        dim: Dimension of the spin-orbital basis.
        na: Number of spin down electrons

    Returns:
        DualBasis representing the cosntraint that is length 1
    """
    dbe = DualBasisElement()
    for i in range(dim // 2):
        dbe.add_element('ck', (2 * i + 1, 2 * i + 1), 1.0)
    dbe.dual_scalar = nb
    return DualBasis(elements=[dbe])


def tpdm_to_thdm_mapping(dim: int) -> DualBasis:
    """
    Generate the dual basis elements for a mapping of the 2-RDM to the
    two-hole-RDM

    Args:
        dim: Dimension of the spin-orbital basis

    Returns:
        DualBasis representing the equality constraint.
    """
    dbe_list = []

    def d2q2element(p: int, q: int, r: int, s: int, factor: Union[float, int])\
            -> DualBasisElement:
        """
        Build the dual basis element for symmetric form of 2-marginal

        :param p: tensor index
        :param q: tensor index
        :param r: tensor index
        :param s: tensor index
        :param factor: scaling coeff for a symmetric constraint
        :return: the dual basis of the mapping
        """
        dbe = DualBasisElement()
        dbe.add_element('cckk', (p, q, r, s), -1.0 * factor)
        dbe.add_element('kkcc', (r, s, p, q), +1.0 * factor)
        if q == s:
            dbe.add_element('ck', (p, r), factor)
        if p == r:
            dbe.add_element('ck', (q, s), factor)
        if q == r:
            dbe.add_element('ck', (p, s), -1. * factor)
        if p == s:
            dbe.add_element('ck', (q, r), -1. * factor)

        dbe.dual_scalar = (
            kronecker_delta(q, s) * kronecker_delta(p, r) -
            kronecker_delta(q, r) * kronecker_delta(p, s)) * factor
        return dbe

    for i, j, k, l in product(range(dim), repeat=4):
        if i * dim + j <= k * dim + l:
            db_element = d2q2element(i, j, k, l, 0.5)
            db_element_2 = d2q2element(k, l, i, j, 0.5)
            dbe_list.append(db_element.join_elements(db_element_2))

    return DualBasis(elements=dbe_list)


def tpdm_to_phdm_mapping(dim: int) -> DualBasis:
    """

    Args:
        dim: Dimension of the spin-orbital basis

    Returns:
        DualBasis representing the equality constraint
    """
    dbe_list = []

    def g2d2map(p: int,
                q: int,
                r: int,
                s: int,
                factor: Optional[Union[float, int]] = 1) -> DualBasisElement:
        """
        Build the dual basis element for a symmetric 2-marginal

        :param p: tensor index
        :param q: tensor index
        :param r: tensor index
        :param s: tensor index
        :param factor: weighting of the element
        :return: the dual basis element
        """
        dbe = DualBasisElement()
        if q == s:
            dbe.add_element('ck', (p, r), -1. * factor)
        dbe.add_element('ckck', (p, s, r, q), 1.0 * factor)
        dbe.add_element('cckk', (p, q, r, s), 1.0 * factor)
        dbe.dual_scalar = 0

        return dbe

    for i, j, k, l in product(range(dim), repeat=4):
        if i * dim + j <= k * dim + l:
            db_element = g2d2map(i, j, k, l, factor=0.5)
            db_element_2 = g2d2map(k, l, i, j, factor=0.5)
            dbe_list.append(db_element.join_elements(db_element_2))

    return DualBasis(elements=dbe_list)


def spin_orbital_linear_constraints(dim: int,
                                    num_alpha: Union[int, float],
                                    num_beta: Union[int, float],
                                    constraint_list: List[str],
                                    sz: Optional[Union[None, float, int]] = None
                                   ) -> DualBasis:
    """
    Construct dual basis constraints for 2-positivity in a spin-orbital
    basis

    Args:
        dim: Dimension of the spin-orbital basis
        num_alpha: number of spin-up electrons.
        num_beta: number of spin-down electrons.
        constraint_list: list of which matrices to include in the 2-pos set.
        sz: target magnetic quantum number.

    Returns:
        DualBasis element.
    """
    N = num_alpha + num_beta
    dual_basis = DualBasis()
    if 'cckk' in constraint_list:
        # Natural constraints on particle-conserving 2-RDMs
        # trace constraint on D2
        dual_basis += tpdm_trace_constraint(dim, N * (N - 1))
        # antisymmetry constraint
        dual_basis += tpdm_antisymmetry_constraint(dim)

    if 'ck' in constraint_list:
        dual_basis += tpdm_to_opdm_mapping(dim, N - 1)
        if sz is not None:
            dual_basis += sz_constraint(dim, sz)

        dual_basis += na_constraint(dim, num_alpha)
        dual_basis += nb_constraint(dim, num_beta)

        if 'kc' in constraint_list:
            dual_basis += opdm_to_ohdm_mapping(dim)

    # d2 -> q2
    if 'kkcc' in constraint_list:
        print('tqdm constraints')
        dual_basis += tpdm_to_thdm_mapping(dim)

    # d2 -> g2
    if "ckck" in constraint_list:
        print('phdm constraints')
        dual_basis += tpdm_to_phdm_mapping(dim)

    return dual_basis


def _coord_generator(i, j, k, l):
    """
    Generator for equivalent spin-orbital indices given a set of four
    spin-orbitals. Indices are in chemist notation so for real-valued chemical
    spinless Hamiltonians the integrals have 8-fold symmetry.

    using this function and iterating over the following:
        i >= j && k >= l && ij >= kl
        where ij = i*(i + 1)/2 + j
              kl = k*(k + 1)/2 + l

    spatial real-values
    i, j, k, l
    j, i, k, l
    i, j, l, k
    j, i, l, k
    k, l, i, j
    k, l, j, i
    l, k, i, j
    l, k, j, i
    """
    unique_set = {(i, j, k, l), (j, i, k, l), (i, j, l, k), (j, i, l, k),
                  (k, l, i, j), (k, l, j, i), (l, k, i, j), (l, k, j, i)}
    for index_element in unique_set:
        yield index_element



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests the code in the examples directory of the git repo."""

import os
import unittest

import numpy
import nbformat


class ExamplesTest(unittest.TestCase):

    def setUp(self):
        """Construct test info"""
        self.testing_folder = os.path.join(
            os.path.dirname(__file__),  # Start at this file's directory.
        )

    def test_example(self):
        """Unit test for examples/performance_benchmark.py."""
        numpy.random.seed(1)

        trial_val = 600
        self.assertEqual(trial_val, 600)

        self.assertTrue(os.path.isdir(self.testing_folder))

    def test_can_run_examples_jupyter_notebooks(self):  # pragma: no cover
        """No coverage on this test because it is not run.
           The test is kept as an example.
        """
        for filename in os.listdir(self.testing_folder):
            if not filename.endswith('.ipynb'):
                continue

            path = os.path.join(self.testing_folder, filename)
            notebook = nbformat.read(path, nbformat.NO_CONVERT)
            state = {}

            for cell in notebook.cells:
                if cell.cell_type == 'code' and not is_matplotlib_cell(cell):
                    try:
                        exec(strip_magics_and_shows(cell.source), state)
                    # coverage: ignore
                    except:
                        print('Failed to run {}.'.format(path))
                        raise


def is_matplotlib_cell(cell):  # pragma: no cover
    return "%matplotlib" in cell.source


def strip_magics_and_shows(text):  # pragma: no cover
    """Remove Jupyter magics and pyplot show commands."""
    lines = [
        line for line in text.split('\n') if not contains_magic_or_show(line)
    ]
    return '\n'.join(lines)


def contains_magic_or_show(line):  # pragma: no cover
    return (line.strip().startswith('%') or 'pyplot.show(' in line or
            'plt.show(' in line)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import random

import numpy as np
import pytest

import openfermion
from openfermion.testing import random_interaction_operator_term


@pytest.mark.parametrize('order,real,seed', [(k, r, random.randrange(2 << 30))
                                             for k in [1, 2, 3, 4, 5]
                                             for r in [0, 1] for _ in range(5)])
def test_random_interaction_operator_term(order, real, seed):
    op = random_interaction_operator_term(order, real, seed)

    assert openfermion.is_hermitian(op)

    assert op.constant == 0
    assert op.one_body_tensor.shape == (order,) * 2
    assert op.two_body_tensor.shape == (order,) * 4

    for tensor in (op.one_body_tensor, op.two_body_tensor):
        for indices in np.argwhere(tensor):
            assert len(set(indices)) == order

    op_2 = random_interaction_operator_term(order, real, seed)
    assert op == op_2

    if order == 1:
        assert op.one_body_tensor != 0
        assert op.two_body_tensor != 0
    elif order == 2:
        assert np.all((op.one_body_tensor == 0) == np.eye(2))
    elif order == 3:
        assert np.all(op.one_body_tensor == 0)
    elif order == 4:
        assert np.all(op.one_body_tensor == 0)
    else:
        assert np.all(op.one_body_tensor == 0)
        assert np.all(op.two_body_tensor == 0)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for rdm_mapping_functions.py"""
import os
import unittest

import numpy
import h5py
from openfermion.config import DATA_DIRECTORY, THIS_DIRECTORY
from openfermion.chem import MolecularData
from openfermion.utils.rdm_mapping_functions import (
    kronecker_delta, map_two_pdm_to_two_hole_dm, map_two_pdm_to_one_pdm,
    map_one_pdm_to_one_hole_dm, map_one_hole_dm_to_one_pdm,
    map_two_pdm_to_particle_hole_dm, map_two_hole_dm_to_two_pdm,
    map_two_hole_dm_to_one_hole_dm, map_particle_hole_dm_to_one_pdm,
    map_particle_hole_dm_to_two_pdm)


class RDMMappingTest(unittest.TestCase):

    def setUp(self):
        # load files and marginals from testing folder
        tqdm_h2_sto3g = os.path.join(THIS_DIRECTORY,
                                     'testing/tqdm_H2_sto-3g_singlet_1.4.hdf5')
        with h5py.File(tqdm_h2_sto3g, 'r') as fid:
            self.tqdm_h2_sto3g = fid['tqdm'][...]

        phdm_h2_sto3g = os.path.join(THIS_DIRECTORY,
                                     'testing/phdm_H2_sto-3g_singlet_1.4.hdf5')
        with h5py.File(phdm_h2_sto3g, 'r') as fid:
            self.phdm_h2_sto3g = fid['phdm'][...]

        tqdm_h2_6_31g = os.path.join(THIS_DIRECTORY,
                                     'testing/tqdm_H2_6-31g_singlet_0.75.hdf5')
        with h5py.File(tqdm_h2_6_31g, 'r') as fid:
            self.tqdm_h2_6_31g = fid['tqdm'][...]

        phdm_h2_6_31g = os.path.join(THIS_DIRECTORY,
                                     'testing/phdm_H2_6-31g_singlet_0.75.hdf5')
        with h5py.File(phdm_h2_6_31g, 'r') as fid:
            self.phdm_h2_6_31g = fid['phdm'][...]

        tqdm_lih_sto3g = os.path.join(
            THIS_DIRECTORY, 'testing/tqdm_H1-Li1_sto-3g_singlet_1.45.hdf5')
        with h5py.File(tqdm_lih_sto3g, 'r') as fid:
            self.tqdm_lih_sto3g = fid['tqdm'][...]

        phdm_lih_sto3g = os.path.join(
            THIS_DIRECTORY, 'testing/phdm_H1-Li1_sto-3g_singlet_1.45.hdf5')
        with h5py.File(phdm_lih_sto3g, 'r') as fid:
            self.phdm_lih_sto3g = fid['phdm'][...]

    def test_kronecker_delta_00(self):
        assert kronecker_delta(0, 0) == 1

    def test_kronecker_delta_01(self):
        assert kronecker_delta(0, 1) == 0

    def test_kronecker_delta_10(self):
        assert kronecker_delta(1, 0) == 0

    def test_kronecker_delta_11(self):
        assert kronecker_delta(1, 1) == 1

    def test_kronecker_delta_nonunit_args(self):
        assert kronecker_delta(3, 3) == 1

    def test_tpdm_to_opdm(self):
        # for all files in datadirectory check if this map holds
        for file in filter(lambda x: x.endswith(".hdf5"),
                           os.listdir(DATA_DIRECTORY)):
            molecule = MolecularData(
                filename=os.path.join(DATA_DIRECTORY, file))
            if (molecule.fci_one_rdm is not None and
                    molecule.fci_two_rdm is not None):
                test_opdm = map_two_pdm_to_one_pdm(molecule.fci_two_rdm,
                                                   molecule.n_electrons)
                assert numpy.allclose(test_opdm, molecule.fci_one_rdm)

    def test_opdm_to_oqdm(self):
        for file in filter(lambda x: x.endswith(".hdf5"),
                           os.listdir(DATA_DIRECTORY)):
            molecule = MolecularData(
                filename=os.path.join(DATA_DIRECTORY, file))
            if molecule.fci_one_rdm is not None:
                test_oqdm = map_one_pdm_to_one_hole_dm(molecule.fci_one_rdm)
                true_oqdm = numpy.eye(molecule.n_qubits) - molecule.fci_one_rdm
                assert numpy.allclose(test_oqdm, true_oqdm)

    def test_oqdm_to_opdm(self):
        for file in filter(lambda x: x.endswith(".hdf5"),
                           os.listdir(DATA_DIRECTORY)):
            molecule = MolecularData(
                filename=os.path.join(DATA_DIRECTORY, file))
            if molecule.fci_one_rdm is not None:
                true_oqdm = numpy.eye(molecule.n_qubits) - molecule.fci_one_rdm
                test_opdm = map_one_hole_dm_to_one_pdm(true_oqdm)
                assert numpy.allclose(test_opdm, molecule.fci_one_rdm)

    def test_tqdm_conversions_h2_631g(self):
        # construct the 2-hole-RDM for LiH the slow way
        # TODO: speed up this calculation by directly contracting from the wf.
        filename = "H2_6-31g_singlet_0.75.hdf5"
        molecule = MolecularData(
            filename=os.path.join(DATA_DIRECTORY, filename))
        true_tqdm = self.tqdm_h2_6_31g

        test_tqdm = map_two_pdm_to_two_hole_dm(molecule.fci_two_rdm,
                                               molecule.fci_one_rdm)
        assert numpy.allclose(true_tqdm, test_tqdm)

        true_oqdm = numpy.eye(molecule.n_qubits) - molecule.fci_one_rdm
        test_oqdm = map_two_hole_dm_to_one_hole_dm(
            true_tqdm, molecule.n_qubits - molecule.n_electrons)
        assert numpy.allclose(true_oqdm, test_oqdm)

        test_tpdm = map_two_hole_dm_to_two_pdm(true_tqdm, molecule.fci_one_rdm)
        assert numpy.allclose(test_tpdm, molecule.fci_two_rdm)

    def test_tqdm_conversions_h2_sto3g(self):
        filename = "H2_sto-3g_singlet_1.4.hdf5"
        molecule = MolecularData(
            filename=os.path.join(DATA_DIRECTORY, filename))
        true_tqdm = self.tqdm_h2_sto3g

        test_tqdm = map_two_pdm_to_two_hole_dm(molecule.fci_two_rdm,
                                               molecule.fci_one_rdm)
        assert numpy.allclose(true_tqdm, test_tqdm)

        true_oqdm = numpy.eye(molecule.n_qubits) - molecule.fci_one_rdm
        test_oqdm = map_two_hole_dm_to_one_hole_dm(
            true_tqdm, molecule.n_qubits - molecule.n_electrons)
        assert numpy.allclose(true_oqdm, test_oqdm)

        test_tpdm = map_two_hole_dm_to_two_pdm(true_tqdm, molecule.fci_one_rdm)
        assert numpy.allclose(test_tpdm, molecule.fci_two_rdm)

    def test_tqdm_conversions_lih_sto3g(self):
        filename = "H1-Li1_sto-3g_singlet_1.45.hdf5"
        molecule = MolecularData(
            filename=os.path.join(DATA_DIRECTORY, filename))
        true_tqdm = self.tqdm_lih_sto3g

        test_tqdm = map_two_pdm_to_two_hole_dm(molecule.fci_two_rdm,
                                               molecule.fci_one_rdm)
        assert numpy.allclose(true_tqdm, test_tqdm)

        true_oqdm = numpy.eye(molecule.n_qubits) - molecule.fci_one_rdm
        test_oqdm = map_two_hole_dm_to_one_hole_dm(
            true_tqdm, molecule.n_qubits - molecule.n_electrons)
        assert numpy.allclose(true_oqdm, test_oqdm)

        test_tpdm = map_two_hole_dm_to_two_pdm(true_tqdm, molecule.fci_one_rdm)
        assert numpy.allclose(test_tpdm, molecule.fci_two_rdm)

    def test_phdm_conversions_h2_631g(self):
        filename = "H2_6-31g_singlet_0.75.hdf5"
        molecule = MolecularData(
            filename=os.path.join(DATA_DIRECTORY, filename))
        true_phdm = self.phdm_h2_6_31g

        test_phdm = map_two_pdm_to_particle_hole_dm(molecule.fci_two_rdm,
                                                    molecule.fci_one_rdm)
        assert numpy.allclose(test_phdm, true_phdm)

        test_opdm = map_particle_hole_dm_to_one_pdm(true_phdm,
                                                    molecule.n_electrons,
                                                    molecule.n_qubits)
        assert numpy.allclose(test_opdm, molecule.fci_one_rdm)

        test_tpdm = map_particle_hole_dm_to_two_pdm(true_phdm,
                                                    molecule.fci_one_rdm)
        assert numpy.allclose(test_tpdm, molecule.fci_two_rdm)

    def test_phdm_conversions_h2_sto3g(self):
        filename = "H2_sto-3g_singlet_1.4.hdf5"
        molecule = MolecularData(
            filename=os.path.join(DATA_DIRECTORY, filename))
        true_phdm = self.phdm_h2_sto3g

        test_phdm = map_two_pdm_to_particle_hole_dm(molecule.fci_two_rdm,
                                                    molecule.fci_one_rdm)
        assert numpy.allclose(test_phdm, true_phdm)

        test_opdm = map_particle_hole_dm_to_one_pdm(true_phdm,
                                                    molecule.n_electrons,
                                                    molecule.n_qubits)
        assert numpy.allclose(test_opdm, molecule.fci_one_rdm)

        test_tpdm = map_particle_hole_dm_to_two_pdm(true_phdm,
                                                    molecule.fci_one_rdm)
        assert numpy.allclose(test_tpdm, molecule.fci_two_rdm)

    def test_phdm_conversions_lih_sto3g(self):
        filename = "H1-Li1_sto-3g_singlet_1.45.hdf5"
        molecule = MolecularData(
            filename=os.path.join(DATA_DIRECTORY, filename))
        true_phdm = self.phdm_lih_sto3g

        test_phdm = map_two_pdm_to_particle_hole_dm(molecule.fci_two_rdm,
                                                    molecule.fci_one_rdm)
        assert numpy.allclose(test_phdm, true_phdm)

        test_opdm = map_particle_hole_dm_to_one_pdm(true_phdm,
                                                    molecule.n_electrons,
                                                    molecule.n_qubits)
        assert numpy.allclose(test_opdm, molecule.fci_one_rdm)

        test_tpdm = map_particle_hole_dm_to_two_pdm(true_phdm,
                                                    molecule.fci_one_rdm)
        assert numpy.allclose(test_tpdm, molecule.fci_two_rdm)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Faster commutators for two-body operators with diagonal Coulomb terms."""

import warnings

from openfermion.ops.operators import FermionOperator
from openfermion.transforms.opconversions import term_reordering


def commutator_ordered_diagonal_coulomb_with_two_body_operator(
        operator_a, operator_b, prior_terms=None):
    """Compute the commutator of two-body operators provided that both are
    normal-ordered and that the first only has diagonal Coulomb interactions.

    Args:
        operator_a: The first FermionOperator argument of the commutator.
            All terms must be normal-ordered, and furthermore either hopping
            operators (i^ j) or diagonal Coulomb operators (i^ i or i^ j^ i j).
        operator_b: The second FermionOperator argument of the commutator.
            operator_b can be any arbitrary two-body operator.
        prior_terms (optional): The initial FermionOperator to add to.

    Returns:
        The commutator, or the commutator added to prior_terms if provided.

    Notes:
        The function could be readily extended to the case of arbitrary
        two-body operator_a given that operator_b has the desired form;
        however, the extra check slows it down without desirable added utility.
    """
    if prior_terms is None:
        prior_terms = FermionOperator.zero()

    for term_a in operator_a.terms:
        coeff_a = operator_a.terms[term_a]
        for term_b in operator_b.terms:
            coeff_b = operator_b.terms[term_b]

            coefficient = coeff_a * coeff_b

            # If term_a == term_b the terms commute, nothing to add.
            if term_a == term_b or not term_a or not term_b:
                continue

            # Case 1: both operators are two-body, operator_a is i^ j^ i j.
            if (len(term_a) == len(term_b) == 4 and
                    term_a[0][0] == term_a[2][0] and
                    term_a[1][0] == term_a[3][0]):
                _commutator_two_body_diagonal_with_two_body(
                    term_a, term_b, coefficient, prior_terms)

            # Case 2: commutator of a 1-body and a 2-body operator
            elif (len(term_b) == 4 and
                  len(term_a) == 2) or (len(term_a) == 4 and len(term_b) == 2):
                _commutator_one_body_with_two_body(term_a, term_b, coefficient,
                                                   prior_terms)

            # Case 3: both terms are one-body operators (both length 2)
            elif len(term_a) == 2 and len(term_b) == 2:
                _commutator_one_body_with_one_body(term_a, term_b, coefficient,
                                                   prior_terms)

            # Final case (case 4): violation of the input promise. Still
            # compute the commutator, but warn the user.
            else:
                warnings.warn('Defaulted to standard commutator evaluation '
                              'due to an out-of-spec operator.')
                additional = FermionOperator.zero()
                additional.terms[term_a + term_b] = coefficient
                additional.terms[term_b + term_a] = -coefficient
                additional = term_reordering.normal_ordered(additional)

                prior_terms += additional

    return prior_terms


def _commutator_one_body_with_one_body(one_body_action_a, one_body_action_b,
                                       coefficient, prior_terms):
    """Compute the commutator of two one-body operators specified by actions.

    Args:
        one_body_action_a, one_body_action_b (tuple): single terms of one-body
            FermionOperators (i^ j or i^ i).
        coefficient (complex float): coefficient of the commutator.
        prior_terms (FermionOperator): prior terms to add the commutator to.
    """
    # In the case that both the creation and annihilation operators of the
    # two actions pair, two new terms must be added.
    if (one_body_action_a[0][0] == one_body_action_b[1][0] and
            one_body_action_b[0][0] == one_body_action_a[1][0]):
        new_one_body_action_a = ((one_body_action_a[0][0], 1),
                                 (one_body_action_a[0][0], 0))
        new_one_body_action_b = ((one_body_action_b[0][0], 1),
                                 (one_body_action_b[0][0], 0))

        prior_terms.terms[new_one_body_action_a] = (
            prior_terms.terms.get(new_one_body_action_a, 0.0) + coefficient)
        prior_terms.terms[new_one_body_action_b] = (
            prior_terms.terms.get(new_one_body_action_b, 0.0) - coefficient)

    # A single pairing causes the mixed action a[0]^ b[1] to be added
    elif one_body_action_a[1][0] == one_body_action_b[0][0]:
        action_ab = ((one_body_action_a[0][0], 1), (one_body_action_b[1][0], 0))

        prior_terms.terms[action_ab] = (prior_terms.terms.get(action_ab, 0.0) +
                                        coefficient)

    # The other single pairing adds the mixed action b[0]^ a[1]
    elif one_body_action_a[0][0] == one_body_action_b[1][0]:
        action_ba = ((one_body_action_b[0][0], 1), (one_body_action_a[1][0], 0))

        prior_terms.terms[action_ba] = (prior_terms.terms.get(action_ba, 0.0) -
                                        coefficient)


def _commutator_one_body_with_two_body(action_a, action_b, coefficient,
                                       prior_terms):
    """Compute commutator of action-specified one- and two-body operators.

    Args:
        action_a, action_b (tuple): single terms, one one-body and the other
            two-body, from normal-ordered FermionOperators. It does not matter
            which is one- or two-body so long as only one of each appears.
        coefficient (complex float): coefficient of the commutator.
        prior_terms (FermionOperator): prior terms to add the commutator to.
    """
    # Determine which action is 1-body and which is 2-body.
    # Label the creation and annihilation parts of the two terms.
    if len(action_a) == 4 and len(action_b) == 2:
        one_body_create = action_b[0][0]
        one_body_annihilate = action_b[1][0]
        two_body_create = (action_a[0][0], action_a[1][0])
        two_body_annihilate = (action_a[2][0], action_a[3][0])
        new_action = list(action_a)
        # Flip coefficient because we reversed the commutator's arguments.
        coefficient *= -1
    else:
        one_body_create = action_a[0][0]
        one_body_annihilate = action_a[1][0]
        two_body_create = (action_b[0][0], action_b[1][0])
        two_body_annihilate = (action_b[2][0], action_b[3][0])
        new_action = list(action_b)

    # If both terms are composed of number operators, they commute.
    if one_body_create == one_body_annihilate and (
            two_body_create == two_body_annihilate):
        return

    # If the one-body annihilation is in the two-body creation parts
    if one_body_annihilate in two_body_create:
        new_coeff = coefficient
        new_inner_action = list(new_action)

        # Determine which creation part(s) of the one-body action to use
        if one_body_annihilate == two_body_create[0]:
            new_inner_action[0] = (one_body_create, 1)
        elif one_body_annihilate == two_body_create[1]:
            new_inner_action[1] = (one_body_create, 1)

        # Normal order if necessary
        if new_inner_action[0][0] < new_inner_action[1][0]:
            new_inner_action[0], new_inner_action[1] = (new_inner_action[1],
                                                        new_inner_action[0])
            new_coeff *= -1

        # Add the resulting term.
        if new_inner_action[0][0] > new_inner_action[1][0]:
            prior_terms.terms[tuple(new_inner_action)] = (
                prior_terms.terms.get(tuple(new_inner_action), 0.0) + new_coeff)

    # If the one-body creation is in the two-body annihilation parts
    if one_body_create in two_body_annihilate:
        new_coeff = -coefficient

        # Determine which annihilation part(s) of the one-body action to sub in
        if one_body_create == two_body_annihilate[0]:
            new_action[2] = (one_body_annihilate, 0)
        elif one_body_create == two_body_annihilate[1]:
            new_action[3] = (one_body_annihilate, 0)

        # Normal order if necessary
        if new_action[2][0] < new_action[3][0]:
            new_action[2], new_action[3] = new_action[3], new_action[2]
            new_coeff *= -1

        # Add the resulting term.
        if new_action[2][0] > new_action[3][0]:
            prior_terms.terms[tuple(new_action)] = (
                prior_terms.terms.get(tuple(new_action), 0.0) + new_coeff)


def _commutator_two_body_diagonal_with_two_body(diagonal_coulomb_action,
                                                arbitrary_two_body_action,
                                                coefficient, prior_terms):
    """Compute the commutator of two two-body operators specified by actions.

    Args:
        diagonal_coulomb_action (tuple): single term of a diagonal Coulomb
            FermionOperator (i^ j^ i j). Must be in normal-ordered form,
            i.e. i > j.
        arbitrary_two_body_action (tuple): arbitrary single term of a two-body
            FermionOperator, in normal-ordered form, i.e. i^ j^ k l with
            i > j, k > l.
        coefficient (complex float): coefficient of the commutator.
        prior_terms (FermionOperator): prior terms to add the commutator to.

    Notes:
        The function could be readily extended to the case of reversed input
        arguments (where diagonal_coulomb_action is the arbitrary one, and
        arbitrary_two_body_action is from a diagonal Coulomb FermionOperator);
        however, the extra check slows it down without significantly increased
        utility.
    """
    # Identify creation and annihilation parts of arbitrary_two_body_action.
    arb_2bdy_create = (arbitrary_two_body_action[0][0],
                       arbitrary_two_body_action[1][0])
    arb_2bdy_annihilate = (arbitrary_two_body_action[2][0],
                           arbitrary_two_body_action[3][0])

    # The first two sub-cases cover when the creations and annihilations of
    # diagonal_coulomb_action and arbitrary_two_body_action totally pair up.
    if (diagonal_coulomb_action[2][0] == arbitrary_two_body_action[0][0] and
            diagonal_coulomb_action[3][0] == arbitrary_two_body_action[1][0]):
        prior_terms.terms[arbitrary_two_body_action] = (
            prior_terms.terms.get(arbitrary_two_body_action, 0.0) - coefficient)

    elif (diagonal_coulomb_action[0][0] == arbitrary_two_body_action[2][0] and
          diagonal_coulomb_action[1][0] == arbitrary_two_body_action[3][0]):
        prior_terms.terms[arbitrary_two_body_action] = (
            prior_terms.terms.get(arbitrary_two_body_action, 0.0) + coefficient)

    # Exactly one of diagonal_coulomb_action's creations matches one of
    # arbitrary_two_body_action's annihilations.
    elif diagonal_coulomb_action[0][0] in arb_2bdy_annihilate:
        # Nothing gets added if there's an unbalanced double creation.
        if diagonal_coulomb_action[1][0] in arb_2bdy_create or (
                diagonal_coulomb_action[0][0] in arb_2bdy_create):
            return

        _add_three_body_term(arbitrary_two_body_action, coefficient,
                             diagonal_coulomb_action[1][0], prior_terms)

    elif diagonal_coulomb_action[1][0] in arb_2bdy_annihilate:
        # Nothing gets added if there's an unbalanced double creation.
        if diagonal_coulomb_action[0][0] in arb_2bdy_create or (
                diagonal_coulomb_action[1][0] in arb_2bdy_create):
            return

        _add_three_body_term(arbitrary_two_body_action, coefficient,
                             diagonal_coulomb_action[0][0], prior_terms)

    elif diagonal_coulomb_action[0][0] in arb_2bdy_create:
        _add_three_body_term(arbitrary_two_body_action, -coefficient,
                             diagonal_coulomb_action[1][0], prior_terms)

    elif diagonal_coulomb_action[1][0] in arb_2bdy_create:
        _add_three_body_term(arbitrary_two_body_action, -coefficient,
                             diagonal_coulomb_action[0][0], prior_terms)


def _add_three_body_term(two_body_action, coefficient, mode, prior_terms):
    new_action = list(two_body_action)

    # Insert creation and annihilation operators into the two-body action.
    new_action.insert(0, (mode, 1))
    new_action.insert(3, (mode, 0))

    # Normal order the creation operators of the new action.
    # Each exchange in the action flips the sign of the coefficient.
    if new_action[0][0] < new_action[1][0]:
        new_action[0], new_action[1] = new_action[1], new_action[0]
        coefficient *= -1
        if new_action[1][0] < new_action[2][0]:
            new_action[1], new_action[2] = new_action[2], new_action[1]
            coefficient *= -1

    # Normal order the annihilation operators of the new action.
    # Each exchange in the action flips the sign of the coefficient.
    if new_action[3][0] < new_action[4][0]:
        new_action[3], new_action[4] = new_action[4], new_action[3]
        coefficient *= -1
        if new_action[4][0] < new_action[5][0]:
            new_action[4], new_action[5] = new_action[5], new_action[4]
            coefficient *= -1

    # Add the new normal-ordered term to the prior terms.
    prior_terms.terms[tuple(new_action)] = (
        prior_terms.terms.get(tuple(new_action), 0.0) + coefficient)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""LinearQubitOperator is a linear operator from QubitOperator."""

import functools
import logging
import multiprocessing

import numpy
import numpy.linalg
import scipy.sparse
import scipy.sparse.linalg

from openfermion.utils.operator_utils import count_qubits


class LinearQubitOperatorOptions(object):
    """Options for LinearQubitOperator."""

    def __init__(self, processes=10, pool=None):
        """
        Args:
            processes(int): Number of processors to use.
            pool(multiprocessing.Pool): A pool of workers.
        """
        if processes <= 0:
            raise ValueError(
                'Invalid number of processors specified {} <= 0'.format(
                    processes))

        self.processes = min(processes, multiprocessing.cpu_count())
        self.pool = pool

    def get_processes(self, num):
        """Number of real processes to use."""
        return max(min(num, self.processes), 1)

    def get_pool(self, num=None):
        """Gets a pool of workers to do some parallel work.

        pool will be cached, which implies that one should be very clear how
        many processes one needs, as it's allocated at most once. Subsequent
        calls of get_pool() will reuse the cached pool.

        Args:
            num(int): Number of workers one needs.
        Returns:
            pool(multiprocessing.Pool): A pool of workers.
        """
        processes = self.get_processes(num or self.processes)
        logging.info("Calling multiprocessing.Pool(%d)", processes)
        return multiprocessing.Pool(processes)


class LinearQubitOperator(scipy.sparse.linalg.LinearOperator):
    """A LinearOperator implied from a QubitOperator.

    The idea is that a single i_th qubit operator, O_i, is a 2-by-2 matrix, to
    be applied on a vector of length n_hilbert / 2^i, performs permutations and/
    or adds an extra factor for its first half and the second half, e.g. a `Z`
    operator keeps the first half unchanged, while adds a factor of -1 to the
    second half, while an `I` keeps it both components unchanged.

    Note that the vector length is n_hilbert / 2^i, therefore when one works on
    i monotonically (in increasing order), one keeps splitting the vector to the
    right size and then apply O_i on them independently.

    Also note that operator O_i, is an *envelop operator* for all operators
    after it, i.e. {O_j | j > i}, which implies that starting with i = 0, one
    can split the vector, apply O_i, split the resulting vector (cached) again
    for the next operator."""

    def __init__(self, qubit_operator, n_qubits=None):
        """
        Args:
            qubit_operator(QubitOperator): A qubit operator to be applied on
                vectors.
            n_qubits(int): The total number of qubits
        """
        calculated_n_qubits = count_qubits(qubit_operator)
        if n_qubits is None:
            n_qubits = calculated_n_qubits
        elif n_qubits < calculated_n_qubits:
            raise ValueError('Invalid number of qubits specified '
                             '{} < {}.'.format(n_qubits, calculated_n_qubits))

        n_hilbert = 2**n_qubits
        super(LinearQubitOperator, self).__init__(shape=(n_hilbert, n_hilbert),
                                                  dtype=complex)
        self.qubit_operator = qubit_operator
        self.n_qubits = n_qubits

    def _matvec(self, x):
        """Matrix-vector multiplication for the LinearQubitOperator class.

        Args:
          x(numpy.ndarray): 1D numpy array.

        Returns:
          retvec(numpy.ndarray): same to the shape of input vector of x.
        """
        retvec = numpy.zeros(x.shape, dtype=complex)
        # Loop through the terms.
        for qubit_term in self.qubit_operator.terms:
            vecs = [x]
            tensor_factor = 0
            coefficient = self.qubit_operator.terms[qubit_term]

            for pauli_operator in qubit_term:
                # Split vector by half and half for each bit.
                if pauli_operator[0] > tensor_factor:
                    vecs = [
                        v for iter_v in vecs for v in numpy.split(
                            iter_v, 2**(pauli_operator[0] - tensor_factor))
                    ]

                # Note that this is to make sure that XYZ operations always work
                # on vector pairs.
                vec_pairs = [numpy.split(v, 2) for v in vecs]

                # There is an non-identity op here, transform the vector.
                xyz = {
                    'X': lambda vps: [[vp[1], vp[0]] for vp in vps],
                    'Y': lambda vps: [[-1j * vp[1], 1j * vp[0]] for vp in vps],
                    'Z': lambda vps: [[vp[0], -vp[1]] for vp in vps],
                }
                vecs = [
                    v for vp in xyz[pauli_operator[1]](vec_pairs) for v in vp
                ]
                tensor_factor = pauli_operator[0] + 1

            # No need to check tensor_factor, i.e. to deal with bits left.
            retvec += coefficient * numpy.concatenate(vecs)
        return retvec


class ParallelLinearQubitOperator(scipy.sparse.linalg.LinearOperator):
    """A LinearOperator from a QubitOperator with multiple processors."""

    def __init__(self, qubit_operator, n_qubits=None, options=None):
        """
        Args:
            qubit_operator(QubitOperator): A qubit operator to be applied on
                vectors.
            n_qubits(int): The total number of qubits
            options(LinearQubitOperatorOptions): Options for the LinearOperator.
        """
        n_qubits = n_qubits or count_qubits(qubit_operator)
        n_hilbert = 2**n_qubits
        super(ParallelLinearQubitOperator,
              self).__init__(shape=(n_hilbert, n_hilbert), dtype=complex)

        self.qubit_operator = qubit_operator
        self.n_qubits = n_qubits
        self.options = options or LinearQubitOperatorOptions()

        self.qubit_operator_groups = list(
            qubit_operator.get_operator_groups(self.options.processes))
        self.linear_operators = [
            LinearQubitOperator(operator, n_qubits)
            for operator in self.qubit_operator_groups
        ]

    def _matvec(self, x):
        """Matrix-vector multiplication for the LinearQubitOperator class.

        Args:
          x(numpy.ndarray): 1D numpy array.

        Returns:
          retvec(numpy.ndarray): same to the shape of input vector of x.
        """
        if not self.linear_operators:
            return numpy.zeros(x.shape)

        pool = self.options.get_pool(len(self.linear_operators))
        vecs = pool.imap_unordered(
            apply_operator,
            [(operator, x) for operator in self.linear_operators])
        pool.close()
        pool.join()
        return functools.reduce(numpy.add, vecs)


def apply_operator(args):
    """Helper funtion to apply opeartor to a vector."""
    operator, vec = args
    return operator * vec


def generate_linear_qubit_operator(qubit_operator, n_qubits=None, options=None):
    """ Generates a LinearOperator from a QubitOperator.

    Args:
        qubit_operator(QubitOperator): A qubit operator to be applied on
            vectors.
        n_qubits(int): The total number of qubits
        options(LinearQubitOperatorOptions): Options for the
            ParallelLinearQubitOperator.
    Returns:
        linear_operator(scipy.sparse.linalg.LinearOperator): A linear operator.
    """
    if options is None:
        linear_operator = LinearQubitOperator(qubit_operator, n_qubits)
    else:
        linear_operator = ParallelLinearQubitOperator(qubit_operator, n_qubits,
                                                      options)
    return linear_operator



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Functions useful for tests."""

import collections
import itertools

import numpy
from scipy.linalg import qr

from openfermion.ops.operators import QubitOperator
from openfermion.ops.representations import (DiagonalCoulombHamiltonian,
                                             QuadraticHamiltonian,
                                             InteractionOperator)


def haar_random_vector(n, seed=None):
    """Generate an n dimensional Haar randomd vector."""
    if seed is not None:
        numpy.random.seed(seed)
    vector = numpy.random.randn(n).astype(complex)
    vector += 1.j * numpy.random.randn(n).astype(complex)
    normalization = numpy.sqrt(vector.dot(numpy.conjugate(vector)))
    return vector / normalization


def random_antisymmetric_matrix(n, real=False, seed=None):
    """Generate a random n x n antisymmetric matrix."""
    if seed is not None:
        numpy.random.seed(seed)

    if real:
        rand_mat = numpy.random.randn(n, n)
    else:
        rand_mat = numpy.random.randn(n, n) + 1.j * numpy.random.randn(n, n)
    antisymmetric_mat = rand_mat - rand_mat.T
    return antisymmetric_mat


def random_hermitian_matrix(n, real=False, seed=None):
    """Generate a random n x n Hermitian matrix."""
    if seed is not None:
        numpy.random.seed(seed)

    if real:
        rand_mat = numpy.random.randn(n, n)
    else:
        rand_mat = numpy.random.randn(n, n) + 1.j * numpy.random.randn(n, n)
    hermitian_mat = rand_mat + rand_mat.T.conj()
    return hermitian_mat


def random_unitary_matrix(n, real=False, seed=None):
    """Obtain a random n x n unitary matrix."""
    if seed is not None:
        numpy.random.seed(seed)

    if real:
        rand_mat = numpy.random.randn(n, n)
    else:
        rand_mat = numpy.random.randn(n, n) + 1.j * numpy.random.randn(n, n)
    Q, _ = qr(rand_mat)
    return Q


def random_qubit_operator(n_qubits=16,
                          max_num_terms=16,
                          max_many_body_order=16,
                          seed=None):
    prng = numpy.random.RandomState(seed)
    op = QubitOperator()
    num_terms = prng.randint(1, max_num_terms + 1)
    for _ in range(num_terms):
        many_body_order = prng.randint(max_many_body_order + 1)
        term = []
        for _ in range(many_body_order):
            index = prng.randint(n_qubits)
            action = prng.choice(('X', 'Y', 'Z'))
            term.append((index, action))
        coefficient = prng.randn()
        op += QubitOperator(term, coefficient)
    return op


def random_diagonal_coulomb_hamiltonian(n_qubits, real=False, seed=None):
    """Generate a random instance of DiagonalCoulombHamiltonian.

    Args:
        n_qubits: The number of qubits
        real: Whether to use only real numbers in the one-body term
    """
    if seed is not None:
        numpy.random.seed(seed)

    one_body = random_hermitian_matrix(n_qubits, real=real)
    two_body = random_hermitian_matrix(n_qubits, real=True)
    constant = numpy.random.randn()
    return DiagonalCoulombHamiltonian(one_body, two_body, constant)


def random_interaction_operator(n_orbitals,
                                expand_spin=False,
                                real=True,
                                seed=None):
    """Generate a random instance of InteractionOperator.

    Args:
        n_orbitals: The number of orbitals.
        expand_spin: Whether to expand each orbital symmetrically into two
            spin orbitals. Note that if this option is set to True, then
            the total number of orbitals will be doubled.
        real: Whether to use only real numbers.
        seed: A random number generator seed.
    """
    if seed is not None:
        numpy.random.seed(seed)

    if real:
        dtype = float
    else:
        dtype = complex

    # The constant has to be real.
    constant = numpy.random.randn()

    # The one-body tensor is a random Hermitian matrix.
    one_body_coefficients = random_hermitian_matrix(n_orbitals, real)

    # Generate random two-body coefficients.
    two_body_coefficients = numpy.zeros(
        (n_orbitals, n_orbitals, n_orbitals, n_orbitals), dtype)
    for p, q, r, s in itertools.product(range(n_orbitals), repeat=4):
        coeff = numpy.random.randn()
        if not real and len(set([p, q, r, s])) >= 3:
            coeff += 1.j * numpy.random.randn()

        # Four point symmetry.
        two_body_coefficients[p, q, r, s] = coeff
        two_body_coefficients[q, p, s, r] = coeff
        two_body_coefficients[s, r, q, p] = coeff.conjugate()
        two_body_coefficients[r, s, p, q] = coeff.conjugate()

        # Eight point symmetry.
        if real:
            two_body_coefficients[r, q, p, s] = coeff
            two_body_coefficients[p, s, r, q] = coeff
            two_body_coefficients[s, p, q, r] = coeff
            two_body_coefficients[q, r, s, p] = coeff

    # If requested, expand to spin orbitals.
    if expand_spin:
        n_spin_orbitals = 2 * n_orbitals

        # Expand one-body tensor.
        one_body_coefficients = numpy.kron(one_body_coefficients, numpy.eye(2))

        # Expand two-body tensor.
        new_two_body_coefficients = numpy.zeros(
            (n_spin_orbitals, n_spin_orbitals, n_spin_orbitals,
             n_spin_orbitals),
            dtype=complex)
        for p, q, r, s in itertools.product(range(n_orbitals), repeat=4):
            coefficient = two_body_coefficients[p, q, r, s]

            # Mixed spin.
            new_two_body_coefficients[2 * p, 2 * q + 1, 2 * r + 1, 2 *
                                      s] = (coefficient)
            new_two_body_coefficients[2 * p + 1, 2 * q, 2 * r, 2 * s +
                                      1] = (coefficient)

            # Same spin.
            new_two_body_coefficients[2 * p, 2 * q, 2 * r, 2 * s] = coefficient
            new_two_body_coefficients[2 * p + 1, 2 * q + 1, 2 * r + 1, 2 * s +
                                      1] = coefficient
        two_body_coefficients = new_two_body_coefficients

    # Create the InteractionOperator.
    interaction_operator = InteractionOperator(constant, one_body_coefficients,
                                               two_body_coefficients)

    return interaction_operator


def random_quadratic_hamiltonian(n_orbitals,
                                 conserves_particle_number=False,
                                 real=False,
                                 expand_spin=False,
                                 seed=None):
    """Generate a random instance of QuadraticHamiltonian.

    Args:
        n_orbitals(int): the number of orbitals
        conserves_particle_number(bool): whether the returned Hamiltonian
            should conserve particle number
        real(bool): whether to use only real numbers
        expand_spin: Whether to expand each orbital symmetrically into two
            spin orbitals. Note that if this option is set to True, then
            the total number of orbitals will be doubled.

    Returns:
        QuadraticHamiltonian
    """
    if seed is not None:
        numpy.random.seed(seed)

    constant = numpy.random.randn()
    chemical_potential = numpy.random.randn()
    hermitian_mat = random_hermitian_matrix(n_orbitals, real)

    if conserves_particle_number:
        antisymmetric_mat = None
    else:
        antisymmetric_mat = random_antisymmetric_matrix(n_orbitals, real)

    if expand_spin:
        hermitian_mat = numpy.kron(hermitian_mat, numpy.eye(2))
        if antisymmetric_mat is not None:
            antisymmetric_mat = numpy.kron(antisymmetric_mat, numpy.eye(2))

    return QuadraticHamiltonian(hermitian_mat, antisymmetric_mat, constant,
                                chemical_potential)


class EqualsTester(object):
    """Tests equality against user-provided disjoint equivalence groups."""

    def __init__(self, test_case):
        self.groups = [(_ClassUnknownToSubjects(),)]
        self.test_case = test_case

    def add_equality_group(self, *group_items):
        """Tries to add a disjoint equivalence group to the equality tester.
        This methods asserts that items within the group must all be equal to
        each other, but not equal to any items in other groups that have been
        or will be added.

        Args:
            *group_items: The items making up the equivalence group.

        Raises:
            AssertError: Items within the group are not equal to each other, or
                items in another group are equal to items within the new group,
                or the items violate the equals-implies-same-hash rule.
        """

        self.test_case.assertIsNotNone(group_items)

        # Check that group items are equivalent to each other.
        for v1, v2 in itertools.product(group_items, repeat=2):
            # Binary operators should always work.
            self.test_case.assertTrue(v1 == v2)
            self.test_case.assertTrue(not v1 != v2)

            # __eq__ and __ne__ should both be correct or not implemented.
            self.test_case.assertTrue(
                hasattr(v1, '__eq__') == hasattr(v1, '__ne__'))
            # Careful: python2 int doesn't have __eq__ or __ne__.
            if hasattr(v1, '__eq__'):
                eq = v1.__eq__(v2)
                ne = v1.__ne__(v2)
                self.test_case.assertIn((eq, ne),
                                        [(True, False), (NotImplemented, False),
                                         (NotImplemented, NotImplemented)])

        # Check that this group's items don't overlap with other groups.
        for other_group in self.groups:
            for v1, v2 in itertools.product(group_items, other_group):
                # Binary operators should always work.
                self.test_case.assertTrue(not v1 == v2)
                self.test_case.assertTrue(v1 != v2)

                # __eq__ and __ne__ should both be correct or not implemented.
                self.test_case.assertTrue(
                    hasattr(v1, '__eq__') == hasattr(v1, '__ne__'))
                # Careful: python2 int doesn't have __eq__ or __ne__.
                if hasattr(v1, '__eq__'):
                    eq = v1.__eq__(v2)
                    ne = v1.__ne__(v2)
                    self.test_case.assertIn((eq, ne),
                                            [(False, True),
                                             (NotImplemented, True),
                                             (NotImplemented, NotImplemented)])

        # Check that group items hash to the same thing, or are all unhashable.
        hashes = [
            hash(v) if isinstance(v, collections.abc.Hashable) else None
            for v in group_items
        ]
        if len(set(hashes)) > 1:
            examples = ((v1, h1, v2, h2)
                        for v1, h1 in zip(group_items, hashes)
                        for v2, h2 in zip(group_items, hashes)
                        if h1 != h2)
            example = next(examples)
            raise AssertionError(
                'Items in the same group produced different hashes. '
                'Example: hash({}) is {} but hash({}) is {}.'.format(*example))

        # Remember this group, to enable disjoint checks vs later groups.
        self.groups.append(group_items)

    def make_equality_pair(self, factory):
        """Tries to add a disjoint (item, item) group to the equality tester.
        Uses the factory method to produce two different objects containing
        equal items. Asserts that the two object are equal, but not equal to
        any items in other groups that have been or will be added. Adds the
        pair as a group.

        Args:
            factory (Callable[[], Any]): A method for producing independent
                copies of an item.

        Raises:
            AssertError: The factory produces items not equal to each other, or
                items in another group are equal to items from the factory, or
                the items violate the equal-implies-same-hash rule.
        """
        self.add_equality_group(factory(), factory())


class _ClassUnknownToSubjects(object):
    """Equality methods should be able to deal with the unexpected."""

    def __eq__(self, other):
        return isinstance(other, _ClassUnknownToSubjects)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(_ClassUnknownToSubjects)


def module_importable(module):
    """Without importing it, returns whether python module is importable.

    Args:
        module (string): Name of module.

    Returns:
        bool

    """
    import sys
    if sys.version_info >= (3, 4):
        from importlib import util
        plug_spec = util.find_spec(module)
    else:
        # Won't enter unless Python<3.4, so ignore for testing
        import pkgutil  # pragma: ignore
        plug_spec = pkgutil.find_loader(module)  # pragma: no cover
    if plug_spec is None:
        return False
    else:
        return True



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


def geometry_from_pubchem(name: str, structure: str = None):
    """Function to extract geometry using the molecule's name from the PubChem
    database. The 'structure' argument can be used to specify which structure
    info to use to extract the geometry. If structure=None, the geometry will
    be constructed based on 3D info, if available, otherwise on 2D (to keep
    backwards compatibility with the times when the argument 'structure'
    was not implemented).

    Args:
        name: a string giving the molecule's name as required by the PubChem
            database.
        structure: a string '2d' or '3d', to specify a specific structure
            information to be retrieved from pubchem. The default is None.
            Recommended value is '3d'.

    Returns:
        geometry: a list of tuples giving the coordinates of each atom with
        distances in Angstrom.
    """
    import pubchempy

    if structure in ['2d', '3d']:
        pubchempy_molecule = pubchempy.get_compounds(name,
                                                     'name',
                                                     record_type=structure)
    elif structure is None:
        # Ideally get the 3-D geometry if available.
        pubchempy_molecule = pubchempy.get_compounds(name,
                                                     'name',
                                                     record_type='3d')

        # If the 3-D geometry isn't available, get the 2-D geometry instead.
        if not pubchempy_molecule:
            pubchempy_molecule = pubchempy.get_compounds(name,
                                                         'name',
                                                         record_type='2d')
    else:
        raise ValueError('Incorrect value for the argument structure=%s' %
                         structure)

    # Check if pubchempy_molecule is an empty list or None
    if not pubchempy_molecule:
        print("Unable to find structure info in the PubChem database"
              "for the specified molecule %s." % name)
        return None

    pubchempy_geometry = \
        pubchempy_molecule[0].to_dict(properties=['atoms'])['atoms']
    geometry = [(atom['element'], (atom['x'], atom['y'], atom.get('z', 0)))
                for atom in pubchempy_geometry]

    return geometry



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Tests for wave_fitting.py"""

import numpy
from .wave_fitting import prony, fit_known_frequencies


def test_prony_zeros():
    signal = numpy.zeros(10)
    amplitudes, phases = prony(signal)
    assert (len(amplitudes) == 5)
    assert (len(phases) == 5)
    for j in range(5):
        numpy.testing.assert_allclose(amplitudes[j], 0)
        numpy.testing.assert_allclose(phases[j], 0)


def test_prony_signal():
    x_vec = numpy.linspace(0, 1, 11)
    y_vec = (0.5 * numpy.exp(1j * x_vec * 3) + 0.3 * numpy.exp(1j * x_vec * 5) +
             0.15 * numpy.exp(1j * x_vec * 1.5) +
             0.1 * numpy.exp(1j * x_vec * 4) +
             0.05 * numpy.exp(1j * x_vec * 1.2))
    print(y_vec)
    amplitudes, phases = prony(y_vec)
    assert (len(amplitudes) == 5)
    assert (len(phases) == 5)
    for a, p in zip(amplitudes, phases):
        print(a, numpy.angle(p))
    numpy.testing.assert_allclose(numpy.abs(amplitudes[0]), 0.5, atol=1e-4)
    numpy.testing.assert_allclose(numpy.abs(amplitudes[1]), 0.3, atol=1e-4)
    numpy.testing.assert_allclose(numpy.abs(amplitudes[2]), 0.15, atol=1e-4)
    numpy.testing.assert_allclose(numpy.abs(amplitudes[3]), 0.1, atol=1e-4)
    numpy.testing.assert_allclose(numpy.abs(amplitudes[4]), 0.05, atol=1e-4)
    numpy.testing.assert_allclose(numpy.angle(phases[0]), 0.3, atol=1e-4)
    numpy.testing.assert_allclose(numpy.angle(phases[1]), 0.5, atol=1e-4)
    numpy.testing.assert_allclose(numpy.angle(phases[2]), 0.15, atol=1e-4)
    numpy.testing.assert_allclose(numpy.angle(phases[3]), 0.4, atol=1e-4)
    numpy.testing.assert_allclose(numpy.angle(phases[4]), 0.12, atol=1e-4)


def test_fitting_signal():
    frequencies = numpy.array([0.4, 0.5, 0.8])
    amplitudes = numpy.array([0.2, 0.4, 0.4])
    times = numpy.linspace(0, 10, 21)
    signal = numpy.array([
        numpy.sum([
            amp * numpy.exp(1j * time * freq)
            for freq, amp in zip(frequencies, amplitudes)
        ])
        for time in times
    ])
    amplitudes_guess = fit_known_frequencies(signal, times, frequencies)
    assert len(amplitudes_guess == 3)
    for index in range(3):
        assert numpy.isclose(amplitudes_guess[index], amplitudes[index])



import numpy as np
from scipy.sparse import csr_matrix
from openfermion.third_party.representability._dualbasis import \
    DualBasisElement, DualBasis


class TMap(object):

    def __init__(self, tensors):
        """
        provide a map of tensor name to tensors

        :param tensors: list of tensors
        :return: TMap object
        """
        self.tensors = tensors
        self._map = dict(map(lambda x: (x.name, x), self.tensors))

    def __getitem__(self, i):
        return self._map[i]

    def __iter__(self):
        for tt in self.tensors:
            yield tt


class MultiTensor(object):

    def __init__(self, tensors, dual_basis=DualBasis()):
        """
        A collection of tensor objects with maps from name to tensor

        Args:
            tensors: a dictionary or tuple of tensors and their associated call
                     name.
            DualBasisElement dual_basis: the set of linear operators restricting
                                         the feasible elements of the linear
                                         space generated by vectorizing the
                                         tensors.
        """
        if not isinstance(tensors, list):
            raise TypeError("MultiTensor accepts a list")

        # this preserves the order the user passes with the tensors
        self.tensors = TMap(tensors)

        # since all the tensors are indexed from zero...I need to know their
        # numbering offset when combined with everything.
        self.off_set_map = self.make_offset_dict(self.tensors)

        # An iterable object that provides access to the dual basis elements
        self.dual_basis = dual_basis
        self.vec_dim = sum([vec.size for vec in self.tensors])

    @staticmethod
    def make_offset_dict(tensors):
        if not isinstance(tensors, TMap):
            raise TypeError("Tensors must be a TMap")

        tally = 0
        offsets = {}
        # this is why ordered dicts are great.  Remembering orderings
        for tensor_value in tensors:
            offsets[tensor_value.name] = tally
            tally += tensor_value.size

        return offsets

    def vectorize_tensors(self):
        """
        vectorize the tensors

        :returns: a vectorized form of the tensors
        """
        vec = np.empty((0, 1))
        for tensor in self.tensors:
            vec = np.vstack((vec, tensor.vectorize()))
        return vec

    def add_dual_elements(self, dual_element):
        """
        Add  a dual element to the dual basis
        """
        if not isinstance(dual_element, DualBasisElement):
            raise TypeError(
                "dual_element variable needs to be a DualBasisElement type")

        # we should extend TMap to add
        self.dual_basis.elements.extend(dual_element)

    def synthesize_dual_basis(self):
        """
        from the list of maps create a m x n sparse matrix for Ax=b

        where x is the vectorized form of all the tensors.  This would be
        the very last step usually.

        :returns: sparse matrix
        """
        # go throught the dual basis list and synthesize each element
        dual_row_indices = []
        dual_col_indices = []
        dual_data_values = []

        # this forms the b-vector of ax + b = c
        bias_data_values = []
        # this forms the c-vector of ax + b = c
        inner_prod_data_values = []
        for index, dual_element in enumerate(self.dual_basis):
            dcol, dval = self.synthesize_element(dual_element)
            dual_row_indices.extend([index] * len(dcol))
            dual_col_indices.extend(dcol)
            dual_data_values.extend(dval)
            inner_prod_data_values.append(float(dual_element.dual_scalar))
            bias_data_values.append(dual_element.constant_bias)
        sparse_dual_operator = csr_matrix(
            (dual_data_values, (dual_row_indices, dual_col_indices)),
            [index + 1, self.vec_dim])

        sparse_bias_vector = csr_matrix(
            (bias_data_values, (range(index + 1), [0] * (index + 1))),
            [index + 1, 1])

        sparse_innerp_vector = csr_matrix(
            (inner_prod_data_values, (range(index + 1), [0] * (index + 1))),
            [index + 1, 1])

        return sparse_dual_operator, sparse_bias_vector, sparse_innerp_vector

    def synthesize_element(self, element):
        """
        Generate the row index and column index for an element

        :param DualBasisElement element: element of the dual basis to vectorize
        """
        col_idx = []
        data_vals = []
        for tlabel, velement, coeff in element:
            col_idx.append(self.off_set_map[tlabel] +
                           self.tensors[tlabel].index_vectorized(*velement))
            data_vals.append(coeff)

        return col_idx, data_vals



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Mapping RDMs to other RDMs"""
from itertools import product
import numpy


def kronecker_delta(i, j):
    return float(i == j)


def map_two_pdm_to_one_pdm(tpdm, particle_number):
    r"""
    Contract a 2-RDM to a 1-RDM

    Args:
        tpdm (numpy.ndarray): The 2-RDM as a 4-index tensor. Indices follow the
            internal convention of tpdm[p, q, r, s] ==
            :math:`a_{p}^{\dagger}a_{q}^{\dagger}a_{r}a_{s}`.
        particle_number (float): number of particles in the system

    Returns:
        opdm (numpy.ndarray): The 1-RDM contracted from the tpdm.
    """
    return numpy.einsum('prrq', tpdm) / (particle_number - 1)


def map_two_pdm_to_two_hole_dm(tpdm, opdm):
    r"""
    Map from the 2-RDM to the 2-hole-RDM

    Args:
        tpdm (numpy.ndarray): The 2-RDM as a 4-index tensor. Indices follow the
            internal convention of tpdm[p, q, r, s] ==
            :math:`a_{p}^{\dagger}a_{q}^{\dagger}a_{r}a_{s}`.
        opdm (numpy.ndarray): The 1-RDM as a 2-index tensor. Indices follow the
            internal convention of opdm[p, q] ==
            :math:`a_{p}^{\dagger}a_{q}`.

    Returns:
        tqdm (numpy.ndarray): The 2-hole matrix.
    """
    ldim = opdm.shape[0]
    tqdm = numpy.zeros_like(tpdm)
    for p, q, r, s in product(range(ldim), repeat=4):
        term1 = (opdm[p, s] * kronecker_delta(q, r) +
                 opdm[q, r] * kronecker_delta(p, s))
        term2 = -1 * (opdm[q, s] * kronecker_delta(p, r) +
                      opdm[p, r] * kronecker_delta(q, s))
        term3 = (kronecker_delta(q, s) * kronecker_delta(p, r) -
                 kronecker_delta(p, s) * kronecker_delta(q, r))
        tqdm[s, r, q, p] = tpdm[p, q, r, s] - term1 - term2 - term3

    return tqdm


def map_two_hole_dm_to_two_pdm(tqdm, opdm):
    r"""
    Map from the 2-hole-RDM to the 2-RDM

    Args:
        tqdm (numpy.ndarray): The 2-hole-RDM as a 4-index tensor. Indices
            follow the internal convention of tqdm[p, q, r, s] ==
            :math:`a_{p}a_{q}a_{r}^{\dagger}a_{s}^{\dagger}`.
        opdm (numpy.ndarray): The 1-RDM as a 2-index tensor. Indices follow the
            internal convention of opdm[p, q] ==
            :math:`a_{p}^{\dagger}a_{q}`.

    Returns:
        tpdm (numpy.ndarray): The 2-RDM matrix.
    """
    ldim = opdm.shape[0]
    tpdm = numpy.zeros_like(tqdm)
    for p, q, r, s in product(range(ldim), repeat=4):
        term1 = (opdm[p, s] * kronecker_delta(q, r) +
                 opdm[q, r] * kronecker_delta(p, s))
        term2 = -1 * (opdm[q, s] * kronecker_delta(p, r) +
                      opdm[p, r] * kronecker_delta(q, s))
        term3 = (kronecker_delta(q, s) * kronecker_delta(p, r) -
                 kronecker_delta(p, s) * kronecker_delta(q, r))
        tpdm[p, q, r, s] = tqdm[r, s, p, q] + term1 + term2 + term3

    return tpdm


def map_two_hole_dm_to_one_hole_dm(tqdm, hole_number):
    r"""
    Map from 2-hole-RDM to 1-hole-RDM

    Args:
        tqdm (numpy.ndarray): The 2-hole-RDM as a 4-index tensor. Indices
            follow the internal convention of tqdm[p, q, r, s] ==
            :math:`a_{p}a_{q}a_{r}^{\dagger}a_{s}^{\dagger}`.
        hole_number (float): Number of holes in the system.  For chemical
                             systems this is usually the number of spin
                             orbitals minus the number of electrons.

    Returns:
        oqdm (numpy.ndarray): The 1-hole-RDM contracted from the tqdm.
    """
    return numpy.einsum('prrq', tqdm) / (hole_number - 1)


def map_one_pdm_to_one_hole_dm(opdm):
    r"""
    Convert a 1-RDM to a 1-hole-RDM

    Args:
        opdm (numpy.ndarray): The 1-RDM as a 2-index tensor. Indices follow the
            internal convention of opdm[p, q] ==
            :math:`a_{p}^{\dagger}a_{q}`.

    Returns:
        oqdm (numpy.ndarray): the 1-hole-RDM transformed from a 1-RDM.
    """
    identity_matrix = numpy.eye(opdm.shape[0])
    return identity_matrix - opdm


def map_one_hole_dm_to_one_pdm(oqdm):
    r"""
    Convert a 1-hole-RDM to a 1-RDM

    Args:
        oqdm (numpy.ndarray): The 1-hole-RDM as a 2-index tensor. Indices
            follow the internal convention of oqdm[p, q] ==
            :math:`a_{p}a_{q}^{\dagger}`.

    Returns:
        oqdm (numpy.ndarray): the 1-hole-RDM transformed from a 1-RDM.
    """
    identity_matrix = numpy.eye(oqdm.shape[0])
    return identity_matrix - oqdm


def map_two_pdm_to_particle_hole_dm(tpdm, opdm):
    r"""
    Map the 2-RDM to the particle-hole-RDM

    Args:
        tpdm (numpy.ndarray): The 2-RDM as a 4-index tensor. Indices follow the
            internal convention of tpdm[p, q, r, s] ==
            :math:`a_{p}^{\dagger}a_{q}^{\dagger}a_{r}a_{s}`.
        opdm (numpy.ndarray): The 1-RDM as a 2-index tensor. Indices follow the
            internal convention of opdm[p, q] ==
            :math:`a_{p}^{\dagger}a_{q}`.

    Returns:
        phdm (numpy.ndarray): The particle-hole matrix.
    """
    ldim = opdm.shape[0]
    phdm = numpy.zeros_like(tpdm)
    for p, q, r, s in product(range(ldim), repeat=4):
        phdm[p, r, q, s] = (opdm[p, s] * kronecker_delta(q, r) -
                            tpdm[p, q, r, s])

    return phdm


def map_particle_hole_dm_to_two_pdm(phdm, opdm):
    r"""
    Map the 2-RDM to the particle-hole-RDM

    Args:
        phdm (numpy.ndarray): The 2-particle-hole-RDM as a 4-index tensor.
            Indices follow the internal convention of phdm[p, q, r, s] ==
            :math:`a_{p}^{\dagger}a_{q}a_{r}^{\dagger}a_{s}`.
        opdm (numpy.ndarray): The 1-RDM as a 2-index tensor. Indices follow the
            internal convention of opdm[p, q] ==
            :math:`a_{p}^{\dagger}a_{q}`.

    Returns:
        tpdm (numpy.ndarray): The 2-RDM matrix.
    """
    ldim = opdm.shape[0]
    tpdm = numpy.zeros_like(phdm)
    for p, q, r, s in product(range(ldim), repeat=4):
        tpdm[p, q, r, s] = (opdm[p, s] * kronecker_delta(q, r) -
                            phdm[p, r, q, s])

    return tpdm


def map_particle_hole_dm_to_one_pdm(phdm, num_particles, num_basis_functions):
    r"""
    Map the particle-hole-RDM to the 1-RDM

    Args:
        phdm (numpy.ndarray): The 2-particle-hole-RDM as a 4-index tensor.
            Indices follow the internal convention of phdm[p, q, r, s] ==
            :math:`a_{p}^{\dagger}a_{q}a_{r}^{\dagger}a_{s}`.
        num_particles: number of particles in the system.
        num_basis_functions: number of spin-orbitals
            (usually the number of qubits)

    Returns:
        opdm (numpy.ndarray): the 1-RDM transformed from a 1-RDM.
    """
    return numpy.einsum('prrq',
                        phdm) / (num_basis_functions - num_particles + 1)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Unit tests of remove_symmetry_qubits(). Test that it preserves the
    correct ground state energy, and reduces the number of qubits required
    by two.
"""

import unittest

from openfermion.hamiltonians import fermi_hubbard
from openfermion.chem import MolecularData
from openfermion.transforms.opconversions import get_fermion_operator
from openfermion.linalg.sparse_tools import (
    get_sparse_operator, jw_get_ground_state_at_particle_number)
from openfermion.linalg import eigenspectrum
from openfermion.ops.operators import FermionOperator

from openfermion.transforms.opconversions.remove_symmetry_qubits import (
    symmetry_conserving_bravyi_kitaev)


def LiH_sto3g():
    """ Generates the Hamiltonian for LiH in
        the STO-3G basis, at a distance of
        1.45 A.
    """
    geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., 1.45))]
    molecule = MolecularData(geometry, 'sto-3g', 1, description="1.45")
    molecule.load()
    molecular_hamiltonian = molecule.get_molecular_hamiltonian()
    hamiltonian = get_fermion_operator(molecular_hamiltonian)
    num_electrons = molecule.n_electrons
    num_orbitals = 2 * molecule.n_orbitals

    return hamiltonian, num_orbitals, num_electrons


def set_1D_hubbard(x_dim):
    """ Returns a 1D Fermi-Hubbard Hamiltonian.
    """
    y_dim = 1
    tunneling = 1.0
    coulomb = 1.0
    magnetic_field = 0.0
    chemical_potential = 0.0
    periodic = False
    spinless = False

    num_orbitals = 2 * x_dim * y_dim

    hubbard_model = fermi_hubbard(x_dim, y_dim, tunneling, coulomb,
                                  chemical_potential, magnetic_field, periodic,
                                  spinless)

    return hubbard_model, num_orbitals


def number_of_qubits(qubit_hamiltonian, unreduced_orbitals):
    """ Returns the number of qubits that a
        qubit Hamiltonian acts upon.
    """
    max_orbital = 0
    for i in range(unreduced_orbitals):
        for term in qubit_hamiltonian.terms:
            if (i, 'X') in term or (i, 'Y') in term or (i, 'Z') in term:
                max_orbital = max_orbital + 1
                break

    return max_orbital


class ReduceSymmetryQubitsTest(unittest.TestCase):

    # Check whether fermionic and reduced qubit Hamiltonians
    # have the same energy for LiH
    def test_energy_reduce_symmetry_qubits(self):
        # Generate the fermionic Hamiltonians,
        # number of orbitals and number of electrons.
        lih_sto_hamil, lih_sto_numorb, lih_sto_numel = LiH_sto3g()

        # Use test function to reduce the qubits.
        lih_sto_qbt = (symmetry_conserving_bravyi_kitaev(
            lih_sto_hamil, lih_sto_numorb, lih_sto_numel))

        self.assertAlmostEqual(
            eigenspectrum(lih_sto_qbt)[0],
            eigenspectrum(lih_sto_hamil)[0])

    # Check that the qubit Hamiltonian acts on two fewer qubits
    # for LiH.
    def test_orbnum_reduce_symmetry_qubits(self):
        # Generate the fermionic Hamiltonians,
        # number of orbitals and number of electrons.
        lih_sto_hamil, lih_sto_numorb, lih_sto_numel = LiH_sto3g()

        # Use test function to reduce the qubits.
        lih_sto_qbt = (symmetry_conserving_bravyi_kitaev(
            lih_sto_hamil, lih_sto_numorb, lih_sto_numel))

        self.assertEqual(number_of_qubits(lih_sto_qbt, lih_sto_numorb),
                         lih_sto_numorb - 2)

    # Check ValueErrors arise correctly.
    def test_errors_reduce_symmetry_qubits(self):
        # Generate the fermionic Hamiltonians,
        # number of orbitals and number of electrons.
        lih_sto_hamil, lih_sto_numorb, lih_sto_numel = LiH_sto3g()

        with self.assertRaises(ValueError):
            symmetry_conserving_bravyi_kitaev(0, lih_sto_numorb, lih_sto_numel)
        with self.assertRaises(ValueError):
            symmetry_conserving_bravyi_kitaev(lih_sto_hamil, 1.5, lih_sto_numel)
        with self.assertRaises(ValueError):
            symmetry_conserving_bravyi_kitaev(lih_sto_hamil, lih_sto_numorb,
                                              3.6)

    # Check energy is the same for Fermi-Hubbard model.
    def test_hubbard_reduce_symmetry_qubits(self):
        for i in range(4):
            n_sites = i + 2
            n_ferm = n_sites
            hub_hamil, n_orb = set_1D_hubbard(n_sites)

            # Use test function to reduce the qubits.
            hub_qbt = (symmetry_conserving_bravyi_kitaev(
                hub_hamil, n_orb, n_ferm))

            sparse_op = get_sparse_operator(hub_hamil)
            ground_energy, _ = jw_get_ground_state_at_particle_number(
                sparse_op, n_ferm)

            self.assertAlmostEqual(eigenspectrum(hub_qbt)[0], ground_energy)

    # Check if single operator of a Hamiltonian is transformed consistently,
    # e.g. system with 2 particles, 4 spin-orbitals, and want to determine
    # elements of one- and two-particle reduced density matrix
    def test_single_operator(self):
        # Dummy operator acting only on 2 qubits of overall 4-qubit system
        op = FermionOperator("0^ 1^ 1 0") + FermionOperator("1^ 0^ 0 1")
        trafo_op = symmetry_conserving_bravyi_kitaev(op,
                                                     active_fermions=2,
                                                     active_orbitals=4)
        # Check via eigenspectrum -- needs to stay the same
        e_op = eigenspectrum(op)
        e_trafo = eigenspectrum(trafo_op)
        # Check eigenvalues
        self.assertSequenceEqual(e_op.tolist(), e_trafo.tolist())



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for commutator_diagonal_coulomb_operator.py"""

import unittest
import warnings

from openfermion.ops.operators import FermionOperator
from openfermion.transforms.opconversions import get_fermion_operator
from openfermion.hamiltonians import jellium_model
from openfermion.utils import commutator, Grid
from openfermion.transforms.opconversions import normal_ordered
from openfermion.transforms.opconversions\
    .commutator_diagonal_coulomb_operator import (
    commutator_ordered_diagonal_coulomb_with_two_body_operator)
from openfermion.testing.testing_utils import (
    random_diagonal_coulomb_hamiltonian)


class DiagonalHamiltonianCommutatorTest(unittest.TestCase):

    def test_commutator(self):
        operator_a = (
            FermionOperator('0^ 0', 0.3) + FermionOperator('1^ 1', 0.1j) +
            FermionOperator('1^ 0^ 1 0', -0.2) + FermionOperator('1^ 3') +
            FermionOperator('3^ 0') + FermionOperator('3^ 2', 0.017) -
            FermionOperator('2^ 3', 1.99) + FermionOperator('3^ 1^ 3 1', .09) +
            FermionOperator('2^ 0^ 2 0', .126j) + FermionOperator('4^ 2^ 4 2') +
            FermionOperator('3^ 0^ 3 0'))

        operator_b = (
            FermionOperator('3^ 1', 0.7) + FermionOperator('1^ 3', -9.) +
            FermionOperator('1^ 0^ 3 0', 0.1) -
            FermionOperator('3^ 0^ 1 0', 0.11) + FermionOperator('3^ 2^ 3 2') +
            FermionOperator('3^ 1^ 3 1', -1.37) + FermionOperator('4^ 2^ 4 2') +
            FermionOperator('4^ 1^ 4 1') + FermionOperator('1^ 0^ 4 0', 16.7) +
            FermionOperator('1^ 0^ 4 3', 1.67) +
            FermionOperator('4^ 3^ 5 2', 1.789j) +
            FermionOperator('6^ 5^ 4 1', -11.789j))

        reference = normal_ordered(commutator(operator_a, operator_b))
        result = commutator_ordered_diagonal_coulomb_with_two_body_operator(
            operator_a, operator_b)

        diff = result - reference
        self.assertTrue(diff.isclose(FermionOperator.zero()))

    def test_nonstandard_second_arg(self):
        operator_a = (FermionOperator('0^ 0', 0.3) +
                      FermionOperator('1^ 1', 0.1j) +
                      FermionOperator('2^ 0^ 2 0', -0.2) +
                      FermionOperator('2^ 1^ 2 1', -0.2j) +
                      FermionOperator('1^ 3') + FermionOperator('3^ 0') +
                      FermionOperator('4^ 4', -1.4j))

        operator_b = (FermionOperator('4^ 1^ 3 0', 0.1) -
                      FermionOperator('3^ 0^ 1 0', 0.11))

        reference = (FermionOperator('1^ 0^ 1 0', -0.11) +
                     FermionOperator('3^ 0^ 1 0', 0.011j) +
                     FermionOperator('3^ 0^ 3 0', 0.11) +
                     FermionOperator('3^ 2^ 0^ 2 1 0', -0.022j) +
                     FermionOperator('4^ 1^ 3 0', -0.03 - 0.13j) +
                     FermionOperator('4^ 2^ 1^ 3 2 0', -0.02 + 0.02j))

        res = commutator_ordered_diagonal_coulomb_with_two_body_operator(
            operator_a, operator_b)

        self.assertTrue(res.isclose(reference))

    def test_add_to_existing_result(self):
        prior_terms = FermionOperator('0^ 1')
        operator_a = FermionOperator('2^ 1')
        operator_b = FermionOperator('0^ 2')

        commutator_ordered_diagonal_coulomb_with_two_body_operator(
            operator_a, operator_b, prior_terms=prior_terms)

        self.assertTrue(prior_terms.isclose(FermionOperator.zero()))

    def test_integration_jellium_hamiltonian_with_negation(self):
        hamiltonian = normal_ordered(
            jellium_model(Grid(2, 3, 1.), plane_wave=False))

        part_a = FermionOperator.zero()
        part_b = FermionOperator.zero()

        add_to_a_or_b = 0  # add to a if 0; add to b if 1
        for term, coeff in hamiltonian.terms.items():
            # Partition terms in the Hamiltonian into part_a or part_b
            if add_to_a_or_b:
                part_a += FermionOperator(term, coeff)
            else:
                part_b += FermionOperator(term, coeff)
            add_to_a_or_b ^= 1

        reference = normal_ordered(commutator(part_a, part_b))
        result = commutator_ordered_diagonal_coulomb_with_two_body_operator(
            part_a, part_b)

        self.assertTrue(result.isclose(reference))

        negative = commutator_ordered_diagonal_coulomb_with_two_body_operator(
            part_b, part_a)
        result += negative

        self.assertTrue(result.isclose(FermionOperator.zero()))

    def test_no_warning_on_nonstandard_input_second_arg(self):
        with warnings.catch_warnings(record=True) as w:
            operator_a = FermionOperator('3^ 2^ 3 2')
            operator_b = FermionOperator('4^ 3^ 4 1')

            reference = FermionOperator('4^ 3^ 2^ 4 2 1')
            result = (
                commutator_ordered_diagonal_coulomb_with_two_body_operator(
                    operator_a, operator_b))

            self.assertFalse(w)

            # Result should still be correct even though we hit the warning.
            self.assertTrue(result.isclose(reference))

    def test_warning_on_bad_input_first_arg(self):
        with warnings.catch_warnings(record=True) as w:
            operator_a = FermionOperator('4^ 3^ 2 1')
            operator_b = FermionOperator('3^ 2^ 3 2')

            reference = normal_ordered(commutator(operator_a, operator_b))
            result = (
                commutator_ordered_diagonal_coulomb_with_two_body_operator(
                    operator_a, operator_b))

            self.assertTrue(len(w) == 1)
            self.assertIn('Defaulted to standard commutator evaluation',
                          str(w[-1].message))

            # Result should still be correct in this case.
            diff = result - reference
            self.assertTrue(diff.isclose(FermionOperator.zero()))

    def test_integration_random_diagonal_coulomb_hamiltonian(self):
        hamiltonian1 = normal_ordered(
            get_fermion_operator(
                random_diagonal_coulomb_hamiltonian(n_qubits=7)))
        hamiltonian2 = normal_ordered(
            get_fermion_operator(
                random_diagonal_coulomb_hamiltonian(n_qubits=7)))

        reference = normal_ordered(commutator(hamiltonian1, hamiltonian2))
        result = commutator_ordered_diagonal_coulomb_with_two_body_operator(
            hamiltonian1, hamiltonian2)

        self.assertTrue(result.isclose(reference))



from typing import List, Optional, Union, Tuple
import copy


class DualBasisElement(object):
    """
    This object is named after the algebraic dual space or dual vector space

    Given a vector x in a vector space V we define the dual basis element
    any (A, b, c) that satisfies

    Ax + b = c

    This object allows a user to specify a linear operator on a space of tensors
    that is vectorized.  Each tensor has a name.  The user constructs `A' and
    provides b and c.  For example if I have a tensor that happens to be a
    matrix called `M'.  The trace operation is an operator `A' where A[i, i] = 1
    for all i in [1, dim(`M')].
    """

    def __init__(self,
                 *,
                 tensor_names: Optional[Union[None, List[str]]] = None,
                 tensor_elements: Optional[
                     Union[None, List[Tuple[int, ...]]]] = None,
                 tensor_coeffs: Optional[Union[None, List[float]]] = None,
                 bias: Optional[int] = 0,
                 scalar: Optional[int] = 0):
        """
        Define a linear operator on a tensor `A', a bias `b', and a result `c'
        satisfying

        .. math::
            Ax + b = c

        Args:
            tensor_names: Names of tensor subspace to act on.
            tensor_elements: List of tuples of numerical tensor indices.
            tensor_coeffs: List of the coefficients corresponding to values to
                           put in `A' at locations `tensor_elements'.
            bias: a linear shift to apply
            scalar: the resulting scalar when the inner product is taken with
                    some element of the tensors `{x}'.
        """
        if tensor_names is None:
            self.primal_tensors_names = []
        else:
            self.primal_tensors_names = tensor_names

        if tensor_elements is None:
            self.primal_elements = []
        else:
            self.primal_elements = tensor_elements

        if tensor_coeffs is None:
            self.primal_coeffs = []
        else:
            self.primal_coeffs = tensor_coeffs

        self.constant_bias = bias
        self.dual_scalar = scalar

    def add_element(self, name, element, coeff):
        """
        Load an element into the data structure
        """
        if not isinstance(element, tuple):
            raise TypeError("element needs to be a tuple of indices")
        if not isinstance(name, str):
            raise TypeError("name needs to be a string")
        if not isinstance(coeff, (float, int)):
            raise TypeError("coeff needs to be a non-complex numerical value")

        self.primal_elements.append(element)
        self.primal_coeffs.append(coeff)
        self.primal_tensors_names.append(name)

    def join_elements(self, other):
        """
        Join two DualBasisElements together to form another DualBasisElement

        Args:
            DualBasisElement other: to join with self
        """
        if not isinstance(other, DualBasisElement):
            raise TypeError("I can only join two DualBasisElements together")

        dbe = DualBasisElement()
        dbe.primal_tensors_names.extend(other.primal_tensors_names)
        dbe.primal_tensors_names.extend(self.primal_tensors_names)

        dbe.primal_elements.extend(other.primal_elements)
        dbe.primal_elements.extend(self.primal_elements)

        dbe.primal_coeffs.extend(other.primal_coeffs)
        dbe.primal_coeffs.extend(self.primal_coeffs)

        dbe.constant_bias += self.constant_bias + other.constant_bias
        dbe.dual_scalar += self.dual_scalar + other.dual_scalar

        return dbe.simplify()

    def simplify(self):
        """
        Mutate the DualBasisElement so that non-unique terms get summed together
        """
        id_dict = {}
        for tname, telement, tcoeff in zip(self.primal_tensors_names,
                                           self.primal_elements,
                                           self.primal_coeffs):
            id_str = tname + ".".join([str(x) for x in telement])
            if id_str not in id_dict:
                id_dict[id_str] = (tname, telement, tcoeff)
            else:
                id_dict[id_str] = (tname, telement, id_dict[id_str][2] + tcoeff)

        tnames = []
        telements = []
        tcoeffs = []
        for _, el in id_dict.items():
            tnames.append(el[0])
            telements.append(el[1])
            tcoeffs.append(el[2])

        self.primal_coeffs = tcoeffs
        self.primal_tensors_names = tnames
        self.primal_elements = telements

        return self

    def id(self):
        """
        Get the unique string identifier for the dual basis element
        """
        id_str = ""
        for name, element in zip(self.primal_tensors_names,
                                 self.primal_elements):
            id_str += name + "(" + ",".join([repr(x) for x in element]) + ")\t"
        return id_str

    def __iter__(self):
        for t_label, velement, coeff in zip(self.primal_tensors_names,
                                            self.primal_elements,
                                            self.primal_coeffs):
            yield t_label, velement, coeff

    def __add__(self, other):
        if isinstance(other, DualBasisElement):
            return DualBasis(elements=[self, other])
        elif isinstance(other, DualBasis):
            return other + self
        else:
            raise TypeError(
                "DualBasisElement can be added to same type or DualBasis")


class DualBasis(object):

    def __init__(
            self,
            elements: Optional[Union[None, List[DualBasisElement]]] = None):
        """
        A collection of DualBasisElements

        Args:
            elements: (optional) a list of DualBasisElement objects
        """

        if elements is None:
            self.elements = []
        else:
            if all(map(lambda x: isinstance(x, DualBasisElement), elements)):
                self.elements = elements
            else:
                raise TypeError("elements must all be DualBasisElement objects")

    def __iter__(self):
        return self.elements.__iter__()

    def __getitem__(self, index):
        return self.elements[index]

    def __len__(self):
        return len(self.elements)

    def __add__(self, other):
        if isinstance(other, DualBasisElement):
            new_elements = copy.deepcopy(self.elements)
            new_elements.append(other)
            return DualBasis(elements=new_elements)

        elif isinstance(other, DualBasis):
            new_elements = copy.deepcopy(self.elements)
            new_elements.extend(other.elements)
            return DualBasis(elements=new_elements)

        else:
            raise TypeError(
                "DualBasis adds DualBasisElements or DualBasis only")



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for interaction_operator.py."""
import unittest

import itertools

import numpy

from openfermion.ops.representations import InteractionOperator
from openfermion.ops.representations import get_active_space_integrals


class ActiveSpaceIntegralsTest(unittest.TestCase):

    def test_raises_error(self):
        with self.assertRaises(ValueError):
            _ = get_active_space_integrals(None, None, [], [])


class InteractionOperatorTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 5
        self.constant = 0.
        self.one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        self.two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)
        self.interaction_operator = InteractionOperator(self.constant,
                                                        self.one_body,
                                                        self.two_body)

    def test_four_point_iter(self):
        constant = 100.0
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)
        one_body[1, 1] = 10.0
        one_body[2, 3] = 11.0
        one_body[3, 2] = 11.0
        two_body[1, 2, 3, 4] = 12.0
        two_body[2, 1, 4, 3] = 12.0
        two_body[3, 4, 1, 2] = 12.0
        two_body[4, 3, 2, 1] = 12.0
        interaction_operator = InteractionOperator(constant, one_body, two_body)

        want_str = '100.0\n10.0\n11.0\n12.0\n'
        got_str = ''
        for key in interaction_operator.unique_iter(complex_valued=True):
            got_str += '{}\n'.format(interaction_operator[key])
        self.assertEqual(want_str, got_str)

    def test_eight_point_iter(self):
        constant = 100.0
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)
        one_body[1, 1] = 10.0
        one_body[2, 3] = 11.0
        one_body[3, 2] = 11.0
        two_body[1, 2, 3, 4] = 12.0
        two_body[2, 1, 4, 3] = 12.0
        two_body[3, 4, 1, 2] = 12.0
        two_body[4, 3, 2, 1] = 12.0
        two_body[1, 4, 3, 2] = 12.0
        two_body[2, 3, 4, 1] = 12.0
        two_body[3, 2, 1, 4] = 12.0
        two_body[4, 1, 2, 3] = 12.0
        interaction_operator = InteractionOperator(constant, one_body, two_body)

        want_str = '100.0\n10.0\n11.0\n12.0\n'
        got_str = ''
        for key in interaction_operator.unique_iter():
            got_str += '{}\n'.format(interaction_operator[key])
        self.assertEqual(want_str, got_str)

    def test_addition(self):
        interaction_op = InteractionOperator(0, numpy.ones((3, 3)),
                                             numpy.ones((3, 3, 3, 3)))
        summed_op = interaction_op + interaction_op
        self.assertTrue(
            numpy.array_equal(summed_op.one_body_tensor,
                              summed_op.n_body_tensors[(1, 0)]))
        self.assertTrue(
            numpy.array_equal(summed_op.two_body_tensor,
                              summed_op.n_body_tensors[(1, 1, 0, 0)]))

    def test_neg(self):
        interaction_op = InteractionOperator(0, numpy.ones((3, 3)),
                                             numpy.ones((3, 3, 3, 3)))
        neg_interaction_op = -interaction_op
        assert isinstance(neg_interaction_op, InteractionOperator)
        assert neg_interaction_op == InteractionOperator(
            0, -numpy.ones((3, 3)), -numpy.ones((3, 3, 3, 3)))

    def test_zero(self):
        interaction_op = InteractionOperator(0, numpy.ones((3, 3)),
                                             numpy.ones((3, 3, 3, 3)))
        assert InteractionOperator.zero(
            interaction_op.n_qubits) == interaction_op * 0

    def test_projected(self):
        interaction_op = InteractionOperator(1, numpy.ones((3, 3)),
                                             numpy.ones((3, 3, 3, 3)))
        projected_two_body_tensor = numpy.zeros_like(
            interaction_op.two_body_tensor)
        for i in range(3):
            projected_two_body_tensor[(i,) * 4] = 1
        projected_op = InteractionOperator(0, numpy.eye(3),
                                           projected_two_body_tensor)
        assert projected_op == interaction_op.projected(1, exact=True)
        projected_op.constant = 1
        assert projected_op == interaction_op.projected(1, exact=False)

        projected_op.one_body_tensor = numpy.zeros_like(
            interaction_op.one_body_tensor)
        for pq in itertools.product(range(2), repeat=2):
            projected_op.one_body_tensor[pq] = 1
        projected_op.two_body_tensor = numpy.zeros_like(
            interaction_op.two_body_tensor)
        for pqrs in itertools.product(range(2), repeat=4):
            projected_op.two_body_tensor[pqrs] = 1
        assert projected_op == interaction_op.projected((0, 1), exact=False)

        projected_op.constant = 0
        projected_op.one_body_tensor = numpy.zeros_like(
            interaction_op.one_body_tensor)
        projected_op.one_body_tensor[0, 1] = 1
        projected_op.one_body_tensor[1, 0] = 1
        projected_op.two_body_tensor = numpy.zeros_like(
            interaction_op.two_body_tensor)
        for pqrs in itertools.product(range(2), repeat=4):
            if len(set(pqrs)) > 1:
                projected_op.two_body_tensor[pqrs] = 1
        assert projected_op == interaction_op.projected((0, 1), exact=True)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Bravyi-Kitaev transform on fermionic operators."""

from openfermion.ops.operators import QubitOperator
from openfermion.transforms.opconversions.bravyi_kitaev import (inline_product,
                                                                inline_sum)
from openfermion.transforms.opconversions.fenwick_tree import FenwickTree


def bravyi_kitaev_tree(operator, n_qubits=None):
    """Apply the "tree" Bravyi-Kitaev transform.

    Implementation from arxiv:1701.07072

    Note that this implementation is different from the one described in
    arXiv:quant-ph/0003137. In particular, it gives different results
    when the total number of modes is not a power of 2. The one described
    in arXiv:quant-ph/0003137 is the same as the one described in
    arXiv:1208.5986, and it is implemented in OpenFermion under the name
    `bravyi_kitaev`.

    Args:
        operator (openfermion.ops.FermionOperator):
            A FermionOperator to transform.
        n_qubits (int|None):
            Can force the number of qubits in the resulting operator above the
            number that appear in the input operator.

    Returns:
        transformed_operator: An instance of the QubitOperator class.

    Raises:
        ValueError: Invalid number of qubits specified.
    """
    # Compute the number of qubits.
    from openfermion.utils import count_qubits
    if n_qubits is None:
        n_qubits = count_qubits(operator)
    if n_qubits < count_qubits(operator):
        raise ValueError('Invalid number of qubits specified.')

    # Build the Fenwick tree.
    fenwick_tree = FenwickTree(n_qubits)

    # Compute transformed operator.
    transformed_terms = (_transform_operator_term(
        term=term, coefficient=operator.terms[term], fenwick_tree=fenwick_tree)
                         for term in operator.terms)
    return inline_sum(summands=transformed_terms, seed=QubitOperator())


def _transform_operator_term(term, coefficient, fenwick_tree):
    """
    Args:
        term (list[tuple[int, int]]):
            A list of (mode, raising-vs-lowering) ladder operator terms.
        coefficient (float):
        fenwick_tree (FenwickTree):
    Returns:
        QubitOperator:
    """

    # Build the Bravyi-Kitaev transformed operators.
    transformed_ladder_ops = (_transform_ladder_operator(
        ladder_operator, fenwick_tree) for ladder_operator in term)
    return inline_product(factors=transformed_ladder_ops,
                          seed=QubitOperator((), coefficient))


def _transform_ladder_operator(ladder_operator, fenwick_tree):
    """
    Args:
        ladder_operator (tuple[int, int]):
        fenwick_tree (FenwickTree):
    Returns:
        QubitOperator:
    """
    index = ladder_operator[0]

    # Parity set. Set of nodes to apply Z to.
    parity_set = [node.index for node in fenwick_tree.get_parity_set(index)]

    # Update set. Set of ancestors to apply X to.
    ancestors = [node.index for node in fenwick_tree.get_update_set(index)]

    # The C(j) set.
    ancestor_children = [
        node.index for node in fenwick_tree.get_remainder_set(index)
    ]

    # Switch between lowering/raising operators.
    d_coefficient = -.5j if ladder_operator[1] else .5j

    # The fermion lowering operator is given by
    # a = (c+id)/2 where c, d are the majoranas.
    d_majorana_component = QubitOperator((((ladder_operator[0], 'Y'),) + tuple(
        (index, 'Z') for index in ancestor_children) + tuple(
            (index, 'X') for index in ancestors)), d_coefficient)

    c_majorana_component = QubitOperator((((ladder_operator[0], 'X'),) + tuple(
        (index, 'Z') for index in parity_set) + tuple(
            (index, 'X') for index in ancestors)), 0.5)

    return c_majorana_component + d_majorana_component



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Gates that target four qubits."""

from typing import Optional, Union

import numpy as np
import sympy

import cirq
from cirq._compat import proper_repr


class DoubleExcitationGate(cirq.EigenGate):
    """Evolve under ``-|0011⟩⟨1100|`` + h.c. for some time."""

    def __init__(
            self,
            *,  # Forces keyword args.
            exponent: Optional[Union[sympy.Symbol, float]] = None,
            rads: Optional[float] = None,
            degs: Optional[float] = None,
            duration: Optional[float] = None) -> None:
        """Initialize the gate.

        At most one of exponent, rads, degs, or duration may be specified.
        If more are specified, the result is considered ambiguous and an
        error is thrown. If no argument is given, the default value of one
        half-turn is used.

        Args:
            exponent: The exponent angle, in half-turns.
            rads: The exponent angle, in radians.
            degs: The exponent angle, in degrees.
            duration: The exponent as a duration of time.
        """

        if len([1 for e in [exponent, rads, degs, duration] if e is not None
               ]) > 1:
            raise ValueError('Redundant exponent specification. '
                             'Use ONE of exponent, rads, degs, or duration.')

        if duration is not None:
            exponent = 2 * duration / np.pi
        else:
            exponent = cirq.chosen_angle_to_half_turns(half_turns=exponent,
                                                       rads=rads,
                                                       degs=degs)

        super().__init__(exponent=exponent)

    def num_qubits(self):
        return 4

    def _eigen_components(self):
        minus_one_component = np.zeros((16, 16))
        minus_one_component[3, 3] = minus_one_component[12, 12] = 0.5
        minus_one_component[3, 12] = minus_one_component[12, 3] = -0.5

        plus_one_component = np.zeros((16, 16))
        plus_one_component[3, 3] = plus_one_component[12, 12] = 0.5
        plus_one_component[3, 12] = plus_one_component[12, 3] = 0.5

        return [(0, np.diag([1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1])),
                (-1, minus_one_component), (1, plus_one_component)]

    def _apply_unitary_(self,
                        args: cirq.ApplyUnitaryArgs) -> Optional[np.ndarray]:
        if cirq.is_parameterized(self):
            return None
        inner_matrix = cirq.unitary(cirq.rx(-2 * np.pi * self.exponent))
        a = args.subspace_index(0b0011)
        b = args.subspace_index(0b1100)
        return cirq.apply_matrix_to_slices(args.target_tensor,
                                           inner_matrix,
                                           slices=[a, b],
                                           out=args.available_buffer)

    def _with_exponent(self, exponent: Union[sympy.Symbol, float]
                      ) -> 'DoubleExcitationGate':
        return DoubleExcitationGate(exponent=exponent)

    def _decompose_(self, qubits):
        p, q, r, s = qubits

        rq_phase_block = [cirq.Z(q)**0.125, cirq.CNOT(r, q), cirq.Z(q)**-0.125]

        srq_parity_transform = [
            cirq.CNOT(s, r), cirq.CNOT(r, q),
            cirq.CNOT(s, r)
        ]

        phase_parity_block = [[
            rq_phase_block, srq_parity_transform, rq_phase_block
        ]]

        yield cirq.CNOT(r, s)
        yield cirq.CNOT(q, p)
        yield cirq.CNOT(q, r)
        yield cirq.X(q)**-self.exponent
        yield phase_parity_block

        yield cirq.CNOT(p, q)
        yield cirq.X(q)
        yield phase_parity_block
        yield cirq.X(q)**self.exponent
        yield phase_parity_block
        yield cirq.CNOT(p, q)
        yield cirq.X(q)

        yield phase_parity_block
        yield cirq.CNOT(q, p)
        yield cirq.CNOT(q, r)
        yield cirq.CNOT(r, s)

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs
                              ) -> cirq.CircuitDiagramInfo:
        if args.use_unicode_characters:
            wire_symbols = ('⇅', '⇅', '⇵', '⇵')
        else:
            # pylint: disable=anomalous-backslash-in-string
            wire_symbols = (r'/\ \/', r'/\ \/', '\/ /\\', '\/ /\\')
        return cirq.CircuitDiagramInfo(wire_symbols=wire_symbols,
                                       exponent=self._diagram_exponent(args))

    def __repr__(self):
        if self.exponent == 1:
            return 'openfermion.DoubleExcitation'
        return '(openfermion.DoubleExcitation**{})'.format(
            proper_repr(self.exponent))


DoubleExcitation = DoubleExcitationGate()
